"""
    Unused binance API calls.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import hashlib
import hmac
import time
from collections import OrderedDict
from urllib.parse import urlencode

import requests
from requests.auth import AuthBase

# Create custom authentication for Exchange
from blankly.exchanges.interfaces.binance.binance_auth import BinanceAuth


class BinanceExchangeAuth(AuthBase):
    def __init__(self, api_key):
        self.api_key = api_key

    def __call__(self, request):
        request.headers['X-MBX-APIKEY'] = self.api_key.encode("utf-8")
        request.headers['Content-Type'] = 'application/json;charset=utf-8'
        return request


def hmac_encode(message: str, secret_key: str) -> str:
    assert isinstance(message, str)
    assert isinstance(secret_key, str)
    signature = hmac.new(secret_key.encode('utf-8'), message.encode('utf-8'), hashlib.sha256)
    return signature.hexdigest()


class API:
    API_URL = 'https://api.binance.{}/api'
    API_TESTNET_URL = 'https://testnet.binance.vision/api'

    def __init__(self, auth: BinanceAuth, tld: str = '.us', testnet: bool = False):
        self.api_key = auth.keys['API_KEY']
        self.secret_key = auth.keys['API_SECRET']

        self.__auth = BinanceExchangeAuth(self.api_key)
        if not testnet:
            self.__api_url = self.API_URL.format(tld)
        else:
            self.__api_url = self.API_TESTNET_URL
        self.session = self._init_session()

    def _init_session(self):
        session = requests.session()
        session.headers.update({'Accept': 'application/json',
                                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
                                              'Chrome/56.0.2924.87 Safari/537.36',
                                'X-MBX-APIKEY': self.api_key.encode('utf-8')})
        return session

    # TODO implement a static download for the orderbook:
    #  https://api.binance.com/api/v3/depth?symbol=BNBBTC&limit=1000

    def _send_request(self, method, url, signed, params=None, data=None):
        if not params:
            params = OrderedDict()
        else:
            assert isinstance(params, OrderedDict)

        if not data:
            data = OrderedDict()
        else:
            assert isinstance(data, OrderedDict)

        params["timestamp"] = int(time.time() * 1000)

        if signed:
            msg = urlencode(params) + urlencode(data)
            params["signature"] = hmac_encode(msg, self.secret_key)

        response = getattr(self.session, method)(url, params=params, data=data)
        return response.json()

    """
    Wallet Endpoints
    """

    def system_status(self):
        endpoint = '/wapi/v3/systemStatus.html'
        signed = False
        return self._send_request('get', self.__api_url + endpoint, signed)

    def all_coins(self):
        endpoint = "/sapi/v1/capital/config/getall"
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def account_snapshot(self, type):
        endpoint = "/sapi/v1/accountSnapshot"
        params = OrderedDict()
        params["type"] = type
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed, params)

    def disable_fast_withdraw(self):
        endpoint = "/sapi/v1/account/disableFastWithdrawSwitch"
        signed = True
        return self._send_request('post', self.__api_url + endpoint, signed)

    def enable_fast_withdraw(self):
        endpoint = "/sapi/v1/account/enableFastWithdrawSwitch"
        signed = True
        return self._send_request('post', self.__api_url + endpoint, signed)

    def withdraw(self, asset, address, amount, **optional_params):
        endpoint = '/wapi/v3/withdraw.html'
        signed = True
        params = OrderedDict()
        params["asset"] = asset
        params["address"] = address
        params["amount"] = amount
        params.update(optional_params)
        return self._send_request('post', self.__api_url + endpoint, signed, params=params)

    def account_info(self):
        endpoint = '/api/v3/account'
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def withdraw_history(self, **optional_params):
        endpoint = '/wapi/v3/depositHistory.html'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def deposit_address(self, **optional_params):
        endpoint = '/wapi/v3/depositAddress.html'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def account_status(self):
        endpoint = '/wapi/v3/accountStatus.html'
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def trading_status(self):
        endpoint = '/wapi/v3/apiTradingStatus.html'
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def dust_log(self):
        endpoint = '/wapi/v3/userAssetDribbletLog.html'
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def dust_transfer(self):
        endpoint = '/sapi/v1/asset/dust'
        signed = True
        return self._send_request('post', self.__api_url + endpoint, signed)

    def dividend_record(self, **optional_params):
        endpoint = '/sapi/v1/asset/assetDividend'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def asset_detail(self, **optional_params):
        endpoint = '/sapi/v1/asset/assetDetail'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed)

    def trade_fee(self, **optional_params):
        endpoint = '/wapi/v3/tradeFee.html'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed)

    def universal_transfer(self, type, asset, amount):
        endpoint = '/wapi/v3/withdraw.html'
        signed = True
        params = OrderedDict()
        params["type"] = type
        params["asset"] = asset
        params["amount"] = amount
        return self._send_request('post', self.__api_url + endpoint, signed, params=params)

    def query_universal_transfer(self, type, **optional_params):
        endpoint = '/sapi/v1/asset/transfer'
        signed = True
        params = OrderedDict()
        params["type"] = type
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    """
    Market Data Endpoints
    """

    def test_connectivity(self):
        endpoint = '/api/v3/ping'
        signed = False
        return self._send_request('get', self.__api_url + endpoint, signed)

    def check_server_time(self):
        endpoint = '/api/v3/time'
        signed = False
        return self._send_request('get', self.__api_url + endpoint, signed)

    def exchange_information(self, symbol):
        endpoint = '/api/v3/exchangeInfo'
        signed = False
        return self._send_request('get', self.__api_url + endpoint, signed)

    def order_book(self, symbol):
        endpoint = '/api/v3/depth'
        signed = False
        params = OrderedDict()
        params["symbol"] = symbol
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def recent_trades_list(self, symbol):
        endpoint = '/api/v3/trades'
        signed = False
        params = OrderedDict()
        params["symbol"] = symbol
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def old_trade_lookup(self, symbol, **optional_params):
        endpoint = '/api/v3/historicalTrades'
        signed = False
        params = OrderedDict()
        params["symbol"] = symbol
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def compressed_trades_list(self, symbol, **optional_params):
        endpoint = '/api/v3/aggTrades'
        signed = False
        params = OrderedDict()
        params["symbol"] = symbol
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def klines(self, symbol, interval, **optional_params):
        endpoint = '/api/v3/klines'
        signed = False
        params = OrderedDict()
        params["symbol"] = symbol
        params["interval"] = interval
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def current_average_price(self, symbol):
        endpoint = '/api/v3/ticker/24hr'
        signed = False
        params = OrderedDict()
        params["symbol"] = symbol
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def ticker_price_change_statistics(self, **optional_params):
        endpoint = '/api/v3/ticker/24hr'
        signed = False
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def symbol_price_ticker(self, **optional_params):
        endpoint = '/api/v3/ticker/price'
        signed = False
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def symbol_orderbook_ticker(self, **optional_params):
        endpoint = '/api/v3/ticker/bookTicker'
        signed = False
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    """
    spot account/trade endpoints
    """

    def test_new_order(self, symbol, side, type, **optional_params):
        endpoint = '/api/v3/order/test'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params["side"] = side
        params["type"] = type
        params.update(optional_params)
        return self._send_request('post', self.__api_url + endpoint, signed, params=params)

    def new_order(self, symbol, side, type, **optional_params):
        endpoint = '/api/v3/order'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params["side"] = side
        params["type"] = type
        params.update(optional_params)
        return self._send_request('post', self.__api_url + endpoint, signed, params=params)

    def cancel_order(self, symbol, **optional_params):
        endpoint = '/api/v3/order'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params.update(optional_params)
        return self._send_request('delete', self.__api_url + endpoint, signed, params=params)

    def cancel_all_open_orders(self, symbol):
        endpoint = 'api/v3/openOrders'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        return self._send_request('delete', self.__api_url + endpoint, signed, params=params)

    def query_order(self, symbol):
        endpoint = '/api/v3/order'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def current_open_orders(self, **optional_params):
        endpoint = '/api/v3/openOrders'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def all_orders(self, symbol, **optional_params):
        endpoint = '/api/v3/allOrders'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def new_oco(self, symbol, side, quantity, price, stop_price, **optional_params):
        endpoint = '/api/v3/order/oco'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params["side"] = side
        params["quantity"] = quantity
        params["price"] = price
        params["stopPrice"] = stop_price
        params.update(optional_params)
        return self._send_request('post', self.__api_url + endpoint, signed, params=params)

    def cancel_oco(self, symbol, **optional_params):
        endpoint = '/api/v3/orderList'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params.update(optional_params)
        return self._send_request('delete', self.__api_url + endpoint, signed, params=params)

    def query_oco(self, **optional_params):
        endpoint = '/api/v3/orderList'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def query_all_oco(self, **optional_params):
        endpoint = '/api/v3/allOrderList'
        signed = True
        params = OrderedDict()
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)

    def query_open_oco(self):
        endpoint = '/api/v3/openOrderList'
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def account_information(self):
        endpoint = '/api/v3/account'
        signed = True
        return self._send_request('get', self.__api_url + endpoint, signed)

    def account_trade_list(self, symbol, **optional_params):
        endpoint = '/api/v3/myTrades'
        signed = True
        params = OrderedDict()
        params["symbol"] = symbol
        params.update(optional_params)
        return self._send_request('get', self.__api_url + endpoint, signed, params=params)



from _pytest.python import Metafunc
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface

from blankly.exchanges.futures.futures_exchange import FuturesExchange

from tests.new_interface_tests.test_utils import FUTURES_EXCHANGES, SYMBOLS


def gen_id(arg):
    if isinstance(arg, FuturesExchange):
        return arg.get_type()
    elif isinstance(arg, FuturesExchangeInterface):
        return arg.get_exchange_type()
    return arg


def pytest_generate_tests(metafunc: Metafunc):
    if 'futures_interface' in metafunc.fixturenames:
        interfaces = [ex.interface for ex in FUTURES_EXCHANGES]
        if 'symbol' in metafunc.fixturenames:
            vals = [(interface, symbol) for interface in interfaces
                    for symbol in SYMBOLS[interface.get_exchange_type()]]
            metafunc.parametrize('futures_interface,symbol', vals, ids=gen_id)
        else:
            metafunc.parametrize('futures_interface', interfaces, ids=gen_id)




"""
    Inherited object to allow library interaction with a blankly bot
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import copy
from multiprocessing import Manager, Process

from binance.client import Client as Binance_API

import blankly
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_api import API as CoinbaseProAPI
from blankly.exchanges.managers.orderbook_manager import OrderbookManager
from blankly.exchanges.managers.ticker_manager import TickerManager
from blankly.utils.utils import info_print


class BlanklyBot:
    symbol: str
    user_preferences: dict
    exchange_type: str
    initial_state: dict
    # Define the given types for the user
    orderbook_manager: OrderbookManager
    ticker_manager: TickerManager
    interface: ABCExchangeInterface
    coinbase_pro_direct: CoinbaseProAPI
    binance_direct: Binance_API

    def __init__(self):
        """
        Initialize state variables when the bot is created
        """
        # If the start is called again, this will make sure multiple processes don't start
        self.__state = {}
        self.coin = ""
        self.exchange_type = ""
        self.initial_state = {}
        self.user_preferences = {}
        self.symbol = ""
        self.direct_calls = None
        self.process = Process(target=self.setup_process)

    def setup(self, exchange_type, currency_pair, user_preferences, initial_state, interface):
        """
        This function is populated by the exchange.
        Args:
            exchange_type (str): Type of exchange i.e "binance" or "coinbase_pro"
            currency_pair: Identifier for the coin market, i.e "BTC-USD"
            user_preferences: Dictionary with the defined preferences
            initial_state: Information about the account the model is defaulted to running on
            interface: Object for consistent trading on the supported exchanges
        """
        # Shared variables with the processing a manager
        self.initial_state = initial_state
        self.__state = Manager().dict({})
        self.exchange_type = exchange_type
        self.user_preferences = user_preferences
        self.interface = copy.deepcopy(interface)
        # Coin id is the currency and which market its on
        self.symbol = currency_pair
        self.direct_calls = interface.get_calls()
        self.coinbase_pro_direct = self.direct_calls
        self.binance_direct = self.direct_calls

    def is_running(self):
        """
        This function returns the status of the process. This ensures that two processes cannot be started on the same
        account.
        """
        return self.process.is_alive()

    def run(self, args):
        """
        Called when the user starts the model. This is what begins and manages the process.
        """
        # Start the setup process and then call the main
        p = Process(target=self.setup_process, args=(args,))
        self.process = p
        self.process.start()

    def setup_process(self, args):
        """
        Create any objects that need to be process-specific in the other process
        """
        self.ticker_manager = blankly.TickerManager(self.exchange_type, self.symbol)
        self.orderbook_manager = blankly.OrderbookManager(self.exchange_type, self.symbol)
        self.main(args)

    """
    State getters and setters for external understanding of the operation
    """

    def get_state(self):
        try:
            return self.__state.copy()
        except BrokenPipeError:
            raise BrokenPipeError("Broken Pipe. Is the main process still running?")

    def update_state(self, key, value):
        """
        This function adds or removes state values from the state dictionary
        Args:
            key: The key value to be added/changed
            value: The value the key should be set to
        """
        self.__state[key] = value

    def remove_key(self, key):
        """
        This function removes a key from the state variable
        Args:
            key: This specifies the key value to remove from the state dictionary
        """
        self.__state.pop(key)

    def main(self, args):
        info_print("No user-created main function to run, stopping...")



import blankly


def price_event(price, symbol, state: blankly.StrategyState):
    """ This function will give an updated price every 15 seconds from our definition below """
    state.variables.history.append(price)

    rsi = blankly.indicators.rsi(state.variables.history)
    current_position = state.interface.account[state.base_asset].available

    if rsi[-1] < 30 and not current_position:
        # Dollar cost average buy
        buy = blankly.trunc(state.interface.cash / price, state.variables.precision)
        state.interface.market_order(symbol, side='buy', size=buy)

    elif rsi[-1] > 70 and current_position:
        # Sell our position
        state.interface.market_order(symbol, side='sell', size=current_position)


def init(symbol, state: blankly.StrategyState):
    # Download price data to give context to the algo
    state.variables.history = state.interface.history(symbol, to=150, return_as='deque',
                                                      resolution=state.resolution)['close']

    # Get the max precision for this symbol from the API
    increment = next(product['base_increment']
                     for product in state.interface.get_products()
                     if product['symbol'] == symbol)
    state.variables.precision = blankly.utils.increment_to_precision(increment)


if __name__ == "__main__":
    # Authenticate EXCHANGE_NAME strategy
    exchange = blankly.EXCHANGE_CLASS()

    # Use our strategy helper on EXCHANGE_NAME
    strategy = blankly.Strategy(exchange)

    # Run the price event function every time we check for a new price - by default that is 15 seconds
    strategy.add_price_event(price_event, symbol='SYMBOL', resolution='1d', init=init)

    # Start our strategy, or run a backtest if this script is run locally.
    if blankly.is_deployed:
        strategy.start()
    else:
        strategy.backtest(to='1y', initial_values={'USD': 10000})



"""
    Oscillator wrappers
    Copyright (C) 2021 Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any

import numpy as np
import pandas as pd
import tulipy as ti

from blankly.indicators.utils import check_series, convert_to_numpy


def rsi(data: Any, period: int = 14, round_rsi: bool = False, use_series=False) -> np.array:
    """ Implements RSI Indicator """
    if period >= len(data):
        return pd.Series() if use_series else []
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    rsi_values = ti.rsi(data, period)
    if round_rsi:
        rsi_values = np.round(rsi_values, 2)
    return pd.Series(rsi_values) if use_series else rsi_values


def aroon_oscillator(high_data: Any, low_data: Any, period=14, use_series=False):
    if check_series(high_data) or check_series(low_data):
        use_series = True
    high_data = convert_to_numpy(high_data)
    low_data = convert_to_numpy(low_data)
    aroonsc = ti.aroonosc(high_data, low_data, period=period)
    return pd.Series(aroonsc) if use_series else aroonsc


def chande_momentum_oscillator(data, period=14, use_series=False):
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    cmo = ti.cmo(data, period)
    return pd.Series(cmo) if use_series else cmo


def absolute_price_oscillator(data, short_period=12, long_period=26, use_series=False):
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    apo = ti.apo(data, short_period, long_period)
    return pd.Series(apo) if use_series else apo


def percentage_price_oscillator(data, short_period=12, long_period=26, use_series=False):
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    ppo = ti.ppo(data, short_period, long_period)
    return pd.Series(ppo) if use_series else ppo


def stochastic_oscillator(high_data, low_data, close_data, pct_k_period=14, pct_k_slowing_period=3, pct_d_period=3,
                          use_series=False):
    if check_series(high_data) or check_series(low_data) or check_series(close_data):
        use_series = True
    high_data = convert_to_numpy(high_data)
    low_data = convert_to_numpy(low_data)
    close_data = convert_to_numpy(close_data)
    stoch = ti.stoch(high_data, low_data, close_data, pct_k_period, pct_k_slowing_period, pct_d_period)
    return pd.Series(stoch) if use_series else stoch


def stochastic_rsi(data, period=14, smooth_pct_k=3, smooth_pct_d=3):
    """ Calculates Stochoastic RSI Courteous of @lukazbinden
    :param data:
    :param period:
    :param smooth_pct_k:
    :param smooth_pct_d:
    :return:
    """
    # Calculate RSI
    rsi_values = rsi(data, period=period, round_rsi=False)

    # Calculate StochRSI
    rsi_values = pd.Series(rsi_values)
    stochrsi = (rsi_values - rsi_values.rolling(period).min()) / (
                rsi_values.rolling(period).max() - rsi_values.rolling(period).min())
    stochrsi_K = stochrsi.rolling(smooth_pct_k).mean()
    stochrsi_D = stochrsi_K.rolling(smooth_pct_d).mean()

    return round(rsi_values, 2), round(stochrsi_K * 100, 2), round(stochrsi_D * 100, 2)



"""
    Utilities for defining how websockets should integrate
    Copyright (C) 2021 Blankly Finance

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import blankly.utils.utils as utils
from typing import List, Dict, Union, Any


def switch_type(stream):
    if stream == "trades":
        return trade, \
               process_trades, \
               "time,system_time,price,open_24h,volume_24h,low_24h,high_24h,volume_30d,best_bid,best_ask," \
               "last_size\n"
    elif stream == "orderbook":
        return no_callback, \
               no_callback, \
               ""
    else:
        return no_callback, no_callback, ""


def no_callback(message):
    return message


"""
Homogenizes response with uhh coinbase and binance
"""


def process_trades(response: dict) -> Dict[str, Union[float, Any]]:
    output = {
              'trade_id': response.pop('id'),
              'time': utils.epoch_from_iso8601(response['time']),
              'size': response['size'],
              'price': response['price']
              }

    return output



def trade(received):
    line = str(received["time"]) + "," + str(time.time()) + "," + received["price"] + "," + received[
        "open_24h"] + "," + received["volume_24h"] + "," + received["low_24h"] + "," + received[
               "high_24h"] + "," + received["volume_30d"] + "," + received["best_bid"] + "," + received[
               "best_ask"] + "," + received["last_size"] + "\n"
    return line



"""
    Abstraction for creating interval driven user strategies for trading futures.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import typing

from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.futures.futures_strategy_logger import FuturesStrategyLogger
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.frameworks.strategy.strategy_base import StrategyBase
from blankly.frameworks.strategy.strategy import Strategy


class FuturesStrategy(Strategy):
    pass


"""
    alpaca exchange definitions and setup
    Copyright (C) 2021 Arun Annamalai, Emerson Dove, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import alpaca_trade_api

from blankly.exchanges.exchange import Exchange
from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.exchanges.interfaces.alpaca.alpaca_api import create_alpaca_client


class Alpaca(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "alpaca", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'alpaca', ['API_KEY', 'API_SECRET', 'sandbox'])

        sandbox = super().evaluate_sandbox(auth)

        calls = create_alpaca_client(auth, sandbox)

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

    def get_exchange_state(self):
        return self.interface.get_products()

    def get_asset_state(self, symbol):
        return self.interface.get_account(symbol)

    def get_direct_calls(self) -> alpaca_trade_api.REST:
        return self.calls

    def get_market_clock(self):
        return self.calls.get_clock()



"""
    Script for managing the variety of tickers on different exchanges.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import random
import requests

import blankly.utils.utils
from blankly.exchanges.interfaces.alpaca.alpaca_websocket import Tickers as Alpaca_Ticker
from blankly.exchanges.interfaces.binance.binance_websocket import Tickers as Binance_Ticker
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_websocket import Tickers as Coinbase_Pro_Ticker
from blankly.exchanges.interfaces.kucoin.kucoin_websocket import Tickers as Kucoin_Ticker
from blankly.exchanges.interfaces.ftx.ftx_websocket import Tickers as FTX_Ticker
from blankly.exchanges.interfaces.okx.okx_websocket import Tickers as Okx_Ticker

from blankly.exchanges.managers.websocket_manager import WebsocketManager


class TickerManager(WebsocketManager):
    def __init__(self, default_exchange: str, default_symbol: str):
        """
        Create a new manager.
        Args:
            default_exchange: Add an exchange name for the manager to favor
            default_symbol: Add a default currency for the manager to favor
        """
        self.__default_exchange = default_exchange
        if default_exchange == "binance":
            default_symbol = blankly.utils.to_exchange_symbol(default_symbol, "binance").lower()
        elif default_exchange == "alpaca":
            default_symbol = blankly.utils.to_exchange_symbol(default_symbol, "alpaca")
        elif default_exchange == "kucoin":
            default_symbol = blankly.utils.to_exchange_symbol(default_symbol, "kucoin")
        elif default_exchange == "ftx":
            default_symbol = blankly.utils.to_exchange_symbol(default_symbol, "ftx")

        self.__default_symbol = default_symbol

        self.__tickers = {default_exchange: {}}

        # Create abstraction for writing many managers
        super().__init__(self.__tickers, default_symbol, default_exchange)

    """ 
    Manager Functions 
    """

    def create_ticker(self, callback, log: str = None, override_symbol: str = None, override_exchange: str = None,
                      **kwargs):
        """
        Create a ticker on a given exchange.
        Args:
            callback: Callback object for the function. Should be something like self.price_event
            log: Fill this with a path to log the price updates.
            override_symbol: The currency to create a ticker for.
            override_exchange: Override the default exchange.
            kwargs: Any keyword arguments to be passed into the callback besides the first positional message argument
        Returns:
            Direct ticker object
        """

        sandbox_mode = self.preferences['settings']['use_sandbox_websockets']

        exchange_name = self.__default_exchange
        # Ensure the ticker dict has this overridden exchange
        if override_exchange is not None:
            if override_exchange not in self.__tickers.keys():
                self.__tickers[override_exchange] = {}
            # Write this value so it can be used later
            exchange_name = override_exchange

        if exchange_name == "coinbase_pro":
            if override_symbol is None:
                override_symbol = self.__default_symbol

            if sandbox_mode:
                ticker = Coinbase_Pro_Ticker(override_symbol, "ticker", log=log,
                                             websocket_url="wss://ws-feed-public.sandbox.pro.coinbase.com", **kwargs)
            else:
                ticker = Coinbase_Pro_Ticker(override_symbol, "ticker", log=log, **kwargs)

            ticker.append_callback(callback)
            # Store this object
            self.__tickers['coinbase_pro'][override_symbol] = ticker
            return ticker
        elif exchange_name == "binance":
            if override_symbol is None:
                override_symbol = self.__default_symbol

            override_symbol = blankly.utils.to_exchange_symbol(override_symbol, "binance").lower()
            if sandbox_mode:
                ticker = Binance_Ticker(override_symbol,
                                        "aggTrade",
                                        log=log,
                                        websocket_url="wss://testnet.binance.vision/ws", **kwargs)
            else:
                ticker = Binance_Ticker(override_symbol,
                                        "aggTrade",
                                        log=log, **kwargs)
            ticker.append_callback(callback)
            override_symbol = override_symbol.upper()
            self.__tickers['binance'][override_symbol] = ticker
            return ticker

        elif exchange_name == "kucoin":
            if override_symbol is None:
                override_symbol = self.__default_symbol

            request_data = (requests.post('https://api.kucoin.com/api/v1/bullet-public').json())

            override_symbol = blankly.utils.to_exchange_symbol(override_symbol, "kucoin")
            if sandbox_mode:
                base_endpoint = request_data['data']['instanceServers'][0]['endpoint']
                token = request_data['data']['token']
                ticker = Kucoin_Ticker(override_symbol,
                                       "ticker",
                                       log=log,
                                       websocket_url=f"{base_endpoint}/socket.io/?token={token}", **kwargs)
            else:
                base_endpoint = request_data['data']['instanceServers'][0]['endpoint']
                token = request_data['data']['token']
                ticker = Kucoin_Ticker(override_symbol, "ticker",
                                       log=log,
                                       websocket_url=f"{base_endpoint}?token={token}&[connectId="
                                                     f"{random.randint(1, 100000000) * 100000000}]", **kwargs)
            ticker.append_callback(callback)
            self.__tickers['kucoin'][override_symbol] = ticker
        elif exchange_name == "okx":
            if override_symbol is None:
                override_symbol = self.__default_symbol

            if sandbox_mode:
                ticker = Okx_Ticker(override_symbol, "tickers", log=log,
                                    WEBSOCKET_URL="wss://wspap.okx.com:8443/ws/v5/public?brokerId=9999", **kwargs)
            else:
                ticker = Okx_Ticker(override_symbol, "tickers", log=log, **kwargs)

            ticker.append_callback(callback)
            # Store this object
            self.__tickers['okx'][override_symbol] = ticker
            return ticker

        elif exchange_name == "alpaca":
            stream = self.preferences['settings']['alpaca']['websocket_stream']
            if override_symbol is None:
                override_symbol = self.__default_symbol

            override_symbol = blankly.utils.to_exchange_symbol(override_symbol, "alpaca")
            if sandbox_mode:
                ticker = Alpaca_Ticker(override_symbol,
                                       "trades",
                                       log=log,
                                       websocket_url="wss://paper-api.alpaca.markets/stream/v2/{}/".format(stream),
                                       **kwargs)
            else:
                ticker = Alpaca_Ticker(override_symbol,
                                       "trades",
                                       log=log,
                                       websocket_url="wss://stream.data.alpaca.markets/v2/{}/".format(stream),
                                       **kwargs)
            ticker.append_callback(callback)
            self.__tickers['alpaca'][override_symbol] = ticker
            return ticker

        elif exchange_name == "ftx":
            if override_symbol is None:
                override_symbol = self.__default_symbol

            override_symbol = blankly.utils.to_exchange_symbol(override_symbol, "ftx")

            if sandbox_mode:
                raise ValueError("Error: FTX does not have a sandbox mode")
            else:
                ticker = FTX_Ticker(override_symbol, "trades", log=log, **kwargs)

            ticker.append_callback(callback)
            # Store this object
            self.__tickers['ftx'][override_symbol] = ticker
            return ticker

        else:
            print(exchange_name + " ticker not supported, skipping creation")



import dataclasses
import functools
import random
from collections import defaultdict
from typing import List, Optional

from blankly.enums import MarginType, HedgeMode, PositionMode, Side, TimeInForce, ContractType, OrderStatus, OrderType
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtesting_wrapper import BacktestingWrapper
from blankly.exchanges.orders.futures.futures_order import FuturesOrder
from blankly.utils import utils as utils
from copy import deepcopy


class FuturesPaperTradeInterface(FuturesExchangeInterface, BacktestingWrapper):
    interface: FuturesExchangeInterface
    paper_account: dict
    paper_positions: dict
    next_fund: int

    _funding_rate_cache: list

    # mapping of contract names -> quote currency used
    # BTC-PERP -> USDT/USD
    _quote_map: dict

    # orders
    _placed_orders: dict
    _executed_orders: dict
    _canceled_orders: dict
    _funds_held: dict

    # not used, just for funsies
    margin_types: dict

    # needed for backtesting_wrapper:
    traded_assets: list

    @property
    def paper_trade_orders(self) -> list:
        return []

    @property
    def executed_orders(self) -> list:
        return []

    @property
    def canceled_orders(self) -> list:
        return []

    @property
    def market_order_execution_details(self) -> list:
        return []

    def add_asset(self, asset):
        if asset not in self.traded_assets:
            self.traded_assets.append(asset)

    @staticmethod
    def convert_symbol_name(symbol: str):
        base, quote = symbol.split('-')
        return base + '-PERP', quote

    def add_symbol(self, symbol):
        contract_name, quote = self.convert_symbol_name(symbol)
        self._quote_map[contract_name] = quote
        self.add_asset(contract_name)
        self.add_asset(quote)

    def __init__(self, exchange_name: str, interface: FuturesExchangeInterface, account_values: dict = None):
        super().__init__(exchange_name, interface.calls)
        BacktestingWrapper.__init__(self)
        self.interface = interface
        self.next_fund = 0

        self._placed_orders = {}
        self._executed_orders = {}
        self._canceled_orders = {}
        self._funds_held = {}
        self._quote_map = {}
        self._funding_rate_cache = []

        # same functionality as 'real' FuturesExchangeInterface's
        self.paper_account = defaultdict(lambda: {'available': 0, 'hold': 0})
        self.paper_positions = {}
        self.traded_assets = []

        self.margin_type = {}
        self.margin_type_global_default = MarginType.CROSSED
        self.leverage = {}
        self.leverage_global_default = 1

        self.init_exchange()
        self.override_local_account(account_values or {})

    def init_exchange(self):
        self.interface.init_exchange()

    @functools.lru_cache(None)
    def get_products(self, symbol: str = None) -> dict:
        return self.interface.get_products(symbol)

    def get_account(self, symbol: str = None) -> dict:
        acc = deepcopy(self.paper_account)
        if symbol:
            try:
                return acc[symbol]
            except KeyError as e:
                if self.backtesting:
                    raise KeyError(
                        "Symbol not found. This can be caused by an invalid quote currency in backtest.json.", e)
                else:
                    raise KeyError("Symbol not found.", e)
        return acc

    def get_position(self, symbol: str = None) -> Optional[dict]:
        if symbol:
            return self.paper_positions.get(symbol, None)
        return self.paper_positions

    def market_order(self, symbol: str, side: Side, size: float, position: PositionMode = PositionMode.BOTH,
                     reduce_only: bool = False) -> FuturesOrder:
        return self._place_order(OrderType.MARKET, symbol, side, size, 0, position, reduce_only)

    def limit_order(self, symbol: str, side: Side, price: float, size: float,
                    position: PositionMode = PositionMode.BOTH,
                    reduce_only: bool = False, time_in_force: TimeInForce = None) -> FuturesOrder:
        return self._place_order(OrderType.LIMIT, symbol, side, size, price, position, reduce_only)

    def take_profit(self, symbol: str, side: Side, price: float, size: float,
                    position: PositionMode = PositionMode.BOTH) -> FuturesOrder:
        return self._place_order(OrderType.TAKE_PROFIT, symbol, side, size, price, position)

    def stop_loss(self, symbol: str, side: Side, price: float, size: float,
                  position: PositionMode = PositionMode.BOTH) -> FuturesOrder:
        return self._place_order(OrderType.STOP, symbol, side, size, price, position)

    def should_run_order(self, order):
        price = self.get_price(order.symbol)

        if order.type == OrderType.MARKET:
            return True

        if order.type == OrderType.LIMIT:
            if order.side == Side.BUY:
                return price <= order.limit_price
            elif order.side == Side.SELL:
                return price >= order.limit_price

        raise Exception(f'invalid order')

    def set_hedge_mode(self, hedge_mode: HedgeMode):
        if hedge_mode == HedgeMode.HEDGE:
            raise ValueError('Hedge mode not supported')

    def get_hedge_mode(self):
        return HedgeMode.ONEWAY

    def set_leverage(self, leverage: float, symbol: str = None):
        if len(self.paper_positions):
            raise Exception('can\'t set leverage with open positions')
        if symbol:
            self.leverage[symbol] = leverage
        else:
            self.leverage_global_default = leverage

    def get_leverage(self, symbol: str = None) -> float:
        return self.leverage.get(symbol, self.leverage_global_default)

    def set_margin_type(self, symbol: str, type: MarginType):
        if len(self.paper_positions):
            raise Exception('can\'t set margin type with open positions')
        if type == MarginType.ISOLATED:
            raise Exception('isolated margin not supported')
        self.margin_type[symbol] = type

    def get_margin_type(self, symbol: str) -> MarginType:
        return self.margin_type.get(symbol, MarginType.CROSSED)

    def cancel_order(self, symbol: str, order_id: int) -> FuturesOrder:
        try:
            order = self._placed_orders[order_id][1]
            del self._placed_orders[order_id]
            assert order.symbol == symbol
            return dataclasses.replace(order, status=OrderStatus.CANCELED)
        except KeyError:
            raise Exception('order not found')

    def get_open_orders(self, symbol: str = None) -> List[FuturesOrder]:
        return [dataclasses.replace(o) for o in self._placed_orders]

    def get_order(self, symbol: str, order_id: int) -> FuturesOrder:
        try:
            order = self._placed_orders[order_id][1]
            assert order.symbol == symbol
            return dataclasses.replace(order, status=OrderStatus.CANCELED)
        except KeyError:
            raise Exception('order not found')

    def get_funding_rate(self, symbol: str):
        if self.backtesting:
            return self.get_backtesting_funding_rate(symbol)
        else:
            return self.interface.get_funding_rate(symbol)

    def get_backtesting_funding_rate(self, symbol: str):
        # check if cache needs to be refreshed
        # TODO this all need to be moved into events framework once that's done
        time = self.time()
        if not self._funding_rate_cache or self._funding_rate_cache[-1]['time'] < time:
            self._funding_rate_cache = self.get_funding_rate_history(symbol, time, None)
        for i, o in self._funding_rate_cache[:-1].enumerate():
            if o['time'] < time:
                last = o['rate']
                next = self._funding_rate_cache[i + 1]['time']
                return last, next
        print(f'failed to download funding rate at time {time}')
        return 0.0001, time + (time % self.get_funding_rate_resolution())

    def get_price(self, symbol: str) -> float:
        if symbol.endswith('-PERP'):
            quote = self._quote_map[symbol]
            symbol = symbol.split('-')[0] + '-' + quote
        if self.backtesting:
            return self.get_backtesting_price(symbol)
        else:
            return self.interface.get_price(symbol)

    def get_funding_rate_history(self, symbol: str, epoch_start: int, epoch_stop: int) -> list:
        return self.interface.get_funding_rate_history(symbol, epoch_start, epoch_stop)

    def get_funding_rate_resolution(self) -> int:
        return self.interface.get_funding_rate_resolution()

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        return self.interface.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def evaluate_traded_account_assets(self):
        pass

    def override_local_account(self, value_dictionary: dict):
        # if python3.9 we could use the new dict update operator here :(
        self.paper_account.update({
            asset: {'available': value, 'hold': 0} for asset, value in value_dictionary.items()
        })

    def do_funding(self, symbol: str, rate: float):
        # positive rate = longs pay for shorts
        # negative rate = shorts pay for longs
        if symbol not in self.paper_positions:
            return
        position = self.paper_positions[symbol]['size']
        if (rate > 0) == (position > 0):
            # we lose money :(
            # either funding rate is positive and we are long
            # or funding rate is negative and we are short
            m = -1
        else:
            m = 1
        self.paper_positions[symbol]['size'] *= 1 + (abs(rate) * m)

    def calculate_position_value(self, symbol):
        position = self.paper_positions.get(symbol, None)
        if not position:
            return 0
        entry_price = position['exchange_specific']['entry_price']
        current_price = self.get_price(symbol) * abs(position['size'])
        return entry_price + (current_price - entry_price) * self.get_leverage(symbol)

    def evaluate_limits(self):
        self.check_margin_call()
        self.execute_orders()

    def execute_orders(self):
        # TODO reduce_only on limit order "technically" broken
        for id, order_data in list(self._placed_orders.items()):
            is_closing, order = order_data
            order: FuturesOrder

            if not self.should_run_order(order):
                continue

            base, quote = order.symbol.split('-')

            asset_price = order.limit_price or self.get_price(order.symbol)
            # if we held funds for the order, use those orelse just default to current price * ordersize
            held_curr, notional = self._funds_held.pop(id, (quote, asset_price * order.size))

            assert held_curr == quote

            # copy order to _executed_orders
            del self._placed_orders[id]
            self._executed_orders[id] = dataclasses.replace(order, status=OrderStatus.FILLED, price=notional)

            position_size = (self.get_position(order.symbol) or {'size': 0})['size']
            size_diff = order.size * (1 if order.side == Side.BUY else -1)

            position = self.paper_positions.get(order.symbol, None)
            entry_price = position['exchange_specific']['entry_price'] if position else 0
            entry_price += notional
            # we sold off our position
            if is_closing:
                value = self.calculate_position_value(order.symbol)
                self.paper_account[quote]['available'] += value
            else:
                self.paper_account[quote]['hold'] -= notional

            # this overwrites but it's fine, we don't need to update
            new_position = position_size + size_diff
            if new_position:
                self.paper_positions[order.symbol] = {
                    'symbol': order.symbol,
                    'base_asset': base,
                    'quote_asset': quote,
                    'size': new_position,
                    'position': PositionMode.BOTH,
                    'leverage': self.get_leverage(order.symbol),
                    'margin_type': self.get_margin_type(order.symbol),
                    'contract_type': ContractType.PERPETUAL,
                    'exchange_specific': {
                        'entry_price': entry_price
                    }
                }
            else:
                del self.paper_positions[order.symbol]
            self.paper_account[base + '-PERP']['available'] = new_position

    def check_margin_call(self):
        called = []
        for symbol, position in self.paper_positions.items():
            value = self.calculate_position_value(symbol)
            if self.cash + value < 0:
                # placing orders in here fucks with the dictionary
                called.append((symbol, position))

        for symbol, position in called:
            self.market_order(symbol, Side.BUY if position['size'] < 0 else Side.SELL, abs(position['size']),
                              reduce_only=True)

    def _place_order(self, type: OrderType, symbol: str, side: Side, size: float, limit_price: float = 0,
                     positionMode: PositionMode = PositionMode.BOTH, reduce_only: bool = False,
                     time_in_force: TimeInForce = TimeInForce.GTC) -> FuturesOrder:
        self.add_symbol(symbol)
        product = self.get_products(symbol)

        if not product:
            raise ValueError(f'invalid symbol {symbol}')

        if positionMode != PositionMode.BOTH:
            raise ValueError(f'only PositionMode.BOTH is supported for paper trading at this time')

        if size <= 0:
            raise ValueError('cannot place empty order')

        if limit_price < 0 or (limit_price == 0 and type != OrderType.MARKET):
            raise ValueError('invalid limit price')

        # place funds on hold
        if type == OrderType.MARKET:
            price = self.get_price(symbol)
        elif reduce_only:
            raise ValueError('reduce_only only supported for market order')
        else:
            price = limit_price

        position = self.get_position(symbol)
        # close position if we are selling from >0 position or buying <0 position
        # no position == we are never closing
        is_closing = self.is_closing_position(position, side)
        if not position and reduce_only:
            raise Exception('reduce_only set to True but there is no position to reduce')

        if reduce_only and not is_closing:
            raise Exception(f'reduce_only set to True but order side ({side}) is opening or growing position')

        if reduce_only:
            size = min(size, abs(position['size']))

        if is_closing and size > abs(position['size']):
            # we need to close position and then buy/sell extra w/ the extra funds
            # this is weird behavior but it's what binance futures does
            # for now raise exception in backtesting, nobody is going to do this anyways
            # and the workaround is simple, just close your position first
            raise Exception('order size is greater than position, close your position first and then place order')

        base, quote = symbol.split('-')
        acc = self.paper_account[quote]

        fee = self.get_taker_fee()
        if is_closing:
            funds = (price * size) * (1 - fee)
        else:
            funds = (price * size) * (1 + fee)

        order_id = self.gen_order_id()
        if not is_closing:
            if acc['available'] < funds:
                raise Exception('not enough funds')

            acc['available'] -= funds
            acc['hold'] += funds
            self._funds_held[order_id] = (quote, funds)

        order = FuturesOrder(symbol=symbol, id=order_id, size=size, status=OrderStatus.OPEN, type=type,
                             contract_type=ContractType.PERPETUAL, side=side, position=PositionMode.BOTH,
                             limit_price=limit_price,
                             price=0, time_in_force=TimeInForce.GTC, response={}, interface=self.interface)
        self._placed_orders[order_id] = (is_closing, order)

        self.execute_orders()

        return order

    def is_closing_position(self, position, side):
        if not position or position['size'] == 0:
            return False
        if (position['size'] > 0) == (side == Side.SELL):
            return True
        return False

    def gen_order_id(self):
        return random.randrange(10 ** 4, 10 ** 5)

    @functools.lru_cache(None)
    def get_maker_fee(self) -> float:
        return self.interface.get_maker_fee()

    @functools.lru_cache(None)
    def get_taker_fee(self) -> float:
        return self.interface.get_taker_fee()



"""
    Strategy State for FuturesStrategy objects.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.utils.utils import AttributeDict, get_base_asset, get_quote_asset, pretty_print_json
from blankly.frameworks.strategy.futures_strategy import FuturesStrategy


class FuturesStrategyState:
    strategy: FuturesStrategy
    interface: FuturesExchangeInterface
    variables: AttributeDict
    resolution: int
    symbol: str
    base_asset: str
    quote_asset: str

    def __init__(self,
                 strategy,
                 variables: AttributeDict,
                 symbol,
                 resolution: [int, float] = None):
        self.strategy = strategy
        self.variables = variables
        self.resolution = resolution
        self.symbol = symbol
        self.base_asset = get_base_asset(symbol)
        self.quote_asset = get_quote_asset(symbol)

    @property
    def interface(self) -> FuturesExchangeInterface:
        return self.strategy.interface

    @property
    def time(self) -> float:
        return self.strategy.time()

    def __str__(self):
        return f"""Symbol: {self.symbol}
Time: {self.time}
Resolution: {self.resolution}
Variables:
{pretty_print_json(self.variables, actually_print=False)}
"""



"""
    alpaca API Interface Definition
    Copyright (C) 2021  Arun Annamalai, Emerson Dove, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import warnings
from datetime import datetime as dt, timezone

import alpaca_trade_api
import dateparser
import pandas as pd
from alpaca_trade_api.rest import APIError as AlpacaAPIError, TimeFrame
from dateutil import parser

from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.utils import utils as utils
from blankly.utils.exceptions import APIException
from blankly.utils.time_builder import build_minute, time_interval_to_seconds, number_interval_to_string


class AlpacaInterface(ExchangeInterface):
    def __init__(self, exchange_name, authenticated_api):
        self.__unique_assets = None
        super().__init__(exchange_name, authenticated_api, valid_resolutions=[60, 60*5, 60*15,
                                                                              60*60*24])
        assert isinstance(self.calls, alpaca_trade_api.REST)

    def init_exchange(self):
        try:
            account_info = self.calls.get_account()
        except alpaca_trade_api.rest.APIError as e:
            raise APIException(e.__str__() + ". Are you trying to use your normal exchange keys "
                               "while in sandbox mode? \nTry toggling the \'sandbox\' setting "
                               "in your keys.json or check if the keys were input correctly into your "
                               "keys.json.")
        try:
            if account_info['account_blocked']:
                warnings.warn('Your alpaca account is indicated as blocked for trading....')
        except KeyError:
            raise LookupError("alpaca API call failed")

        filtered_assets = []
        products = self.calls.list_assets(status=None, asset_class=None)
        for i in products:
            if i['symbol'] not in filtered_assets and i['status'] != 'inactive':
                filtered_assets.append(i['symbol'])
            else:
                # TODO handle duplicate symbols
                pass
        self.__unique_assets = filtered_assets

    def get_products(self) -> dict:
        """
        [
            {
              "id": "904837e3-3b76-47ec-b432-046db621571b",
              "class": "us_equity",
              "exchange": "NASDAQ",
              "symbol": "AAPL",
              "status": "active",
              "tradable": true,
              "marginable": true,
              "shortable": true,
              "easy_to_borrow": true,
              "fractionable": true
            },
            ...
        ]
        """
        needed = self.needed['get_products']
        assets = self.calls.list_assets(status=None, asset_class=None)

        for asset in assets:
            base_asset = asset['symbol']
            asset['symbol'] = base_asset
            asset['base_asset'] = base_asset
            asset['quote_asset'] = 'USD'
            if asset['fractionable']:
                asset['base_min_size'] = .000000001
                asset['base_increment'] = .000000001
            else:
                asset['base_min_size'] = 1
                asset['base_increment'] = 1
            asset['base_max_size'] = 10000000000

        for i in range(len(assets)):
            assets[i] = utils.isolate_specific(needed, assets[i])

        return assets

    @property
    def cash(self):
        account_dict = self.calls.get_account()
        return float(account_dict['buying_power'])

    @utils.enforce_base_asset
    def get_account(self, symbol=None):
        assert isinstance(self.calls, alpaca_trade_api.REST)

        symbol = super().get_account(symbol)

        positions = self.calls.list_positions()
        positions_dict = utils.AttributeDict({})

        for position in positions:
            curr_symbol = position.pop('symbol')
            positions_dict[curr_symbol] = utils.AttributeDict({
                'available': float(position.pop('qty')),
                'hold': 0.0
            })

        symbols = list(positions_dict.keys())
        # Catch an edge case bug that if there are no positions it won't try to snapshot
        if len(symbols) != 0:
            open_orders = self.calls.list_orders(status='open', symbols=symbols)
            snapshot_price = self.calls.get_snapshots(symbols=symbols)
        else:
            open_orders = []
            snapshot_price = {}

        # now grab the available cash in the account
        account = self.calls.get_account()
        positions_dict['USD'] = utils.AttributeDict({
            'available': float(account['buying_power']),
            'hold': 0.0
        })

        for order in open_orders:
            curr_symbol = order['symbol']
            if order['side'] == 'buy':  # buy orders only affect USD holds
                if order['qty']:  # this case handles qty market buy and limit buy
                    if order['type'] == 'limit':
                        dollar_amt = float(order['qty']) * float(order['limit_price'])
                    elif order['type'] == 'market':
                        dollar_amt = float(order['qty']) * snapshot_price[curr_symbol]['latestTrade']['p']
                    else:  # we don't have support for stop_order, stop_limit_order
                        dollar_amt = 0.0
                else:  # this is the case for notional market buy
                    dollar_amt = float(order['notional'])

                # In this case we don't have to subtract because the buying power is the available money already
                # we just need to add to figure out how much is actually on limits
                # positions_dict['USD']['available'] -= dollar_amt

                # So just add to our hold
                positions_dict['USD']['hold'] += dollar_amt

            else:
                if order['qty']:  # this case handles qty market sell and limit sell
                    qty = float(order['qty'])
                else:  # this is the case for notional market sell, calculate the qty with cash/price
                    qty = float(order['notional']) / snapshot_price[curr_symbol]['latestTrade']['p']

                positions_dict[curr_symbol]['available'] -= qty
                positions_dict[curr_symbol]['hold'] += qty

        # Note that now __unique assets could be uninitialized:
        if self.__unique_assets is None:
            self.init_exchange()

        for i in self.__unique_assets:
            if i not in positions_dict:
                positions_dict[i] = utils.AttributeDict({
                    'available': 0.0,
                    'hold': 0.0
                })

        if symbol is not None:
            if symbol in positions_dict:
                return utils.AttributeDict({
                    'available': float(positions_dict[symbol]['available']),
                    'hold': float(positions_dict[symbol]['hold'])
                })
            else:
                raise KeyError('Symbol not found.')

        if symbol == 'USD':
            return utils.AttributeDict({
                'available': positions_dict['USD']['available'],
                'hold': positions_dict['USD']['hold']
            })

        return positions_dict

    @staticmethod
    def __parse_iso(response):
        try:
            response['created_at'] = parser.isoparse(response['created_at']).timestamp()
        except ValueError as e:
            if str(e) == 'Unused components in ISO string':
                response['created_at'] = parser.parse(response['created_at']).timestamp()
            else:
                raise e

        return response

    @utils.order_protection
    def market_order(self, symbol, side, size) -> MarketOrder:
        assert isinstance(self.calls, alpaca_trade_api.REST)
        needed = self.needed['market_order']

        renames = [
            ['qty', 'size']
        ]

        order = {
            'size': size,
            'side': side,
            'symbol': symbol,
            'type': 'market'
        }
        response = self.calls.submit_order(symbol, side=side, type='market', time_in_force='day', qty=size)

        response = self.__parse_iso(response)

        response = utils.rename_to(renames, response)
        response = utils.isolate_specific(needed, response)
        return MarketOrder(order, response, self)

    @utils.order_protection
    def limit_order(self, symbol: str, side: str, price: float, size: float) -> LimitOrder:
        needed = self.needed['limit_order']

        renames = [
            ['limit_price', 'price'],
            ['qty', 'size']
        ]

        order = {
            'quantity': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'limit'
        }
        response = self.calls.submit_order(symbol,
                                           side=side,
                                           type='limit',
                                           time_in_force='gtc',
                                           qty=size,
                                           limit_price=price)

        response = self.__parse_iso(response)

        response = utils.rename_to(renames, response)
        response = utils.isolate_specific(needed, response)
        response['time_in_force'] = response['time_in_force'].upper()
        return LimitOrder(order, response, self)

    def cancel_order(self, symbol, order_id) -> dict:
        assert isinstance(self.calls, alpaca_trade_api.REST)
        self.calls.cancel_order(order_id)

        # TODO: handle the different response codes
        return {'order_id': order_id}

    def get_open_orders(self, symbol=None):
        assert isinstance(self.calls, alpaca_trade_api.REST)
        if symbol is None:
            orders = self.calls.list_orders(status='open')
        else:
            orders = self.calls.list_orders(status='open', symbols=[symbol])

        for i in range(len(orders)):
            # orders[i] = utils.rename_to(renames, orders[i])
            # if orders[i]['type'] == "limit":
            #     orders[i]['price'] = orders[i]['limit_price']
            # if orders[i]['type'] == "market":
            #     if orders[i]['notional']:
            #         orders[i]['funds'] = orders[i]['notional']
            #     else:
            #         orders[i]['funds'] = orders[i]['notional']
            # orders[i]['created_at'] = parser.isoparse(orders[i]['created_at']).timestamp()
            orders[i] = self.homogenize_order(orders[i])

        return orders

    def get_order(self, symbol, order_id) -> dict:
        assert isinstance(self.calls, alpaca_trade_api.REST)
        order = self.calls.get_order(order_id)
        order = self.homogenize_order(order)
        return order

    # TODO: fix this function
    def homogenize_order(self, order):
        if order['type'] == "limit":
            renames = [
                ["qty", "size"],
                ["limit_price", "price"]
            ]
            order = utils.rename_to(renames, order)

        elif order['type'] == "market":
            if order['notional']:
                renames = [
                    ["notional", "funds"]
                ]
                order = utils.rename_to(renames, order)

            else:  # market order of number of shares
                order['size'] = order.pop('qty')

        # TODO: test stop_limit orders
        elif order['type'] == "stop_limit":
            renames = [
                ["qty", "size"],
            ]
            order = utils.rename_to(renames, order)

        order = self.__parse_iso(order)

        needed = self.choose_order_specificity(order['type'])

        order = utils.isolate_specific(needed, order)
        return order

    def get_fees(self, symbol):
        assert isinstance(self.calls, alpaca_trade_api.REST)
        return {
            'maker_fee_rate': 0.0,
            'taker_fee_rate': 0.0
        }

    def get_product_history(self, symbol: str, epoch_start: float, epoch_stop: float, resolution: int):
        if not self.user_preferences['settings']['alpaca']['use_yfinance']:
            assert isinstance(self.calls, alpaca_trade_api.REST)

            resolution = time_interval_to_seconds(resolution)

            supported_multiples = [60, 3600, 86400]
            if resolution not in supported_multiples:
                utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
                resolution = supported_multiples[min(range(len(supported_multiples)),
                                                 key=lambda i: abs(supported_multiples[i] - resolution))]

            found_multiple, row_divisor = super().evaluate_multiples(supported_multiples, resolution)

            if found_multiple == 60:
                time_interval = TimeFrame.Minute
            elif found_multiple == 3600:
                time_interval = TimeFrame.Hour
            else:
                time_interval = TimeFrame.Day

            epoch_start_str = dt.fromtimestamp(epoch_start, tz=timezone.utc).isoformat()
            epoch_stop_str = dt.fromtimestamp(epoch_stop, tz=timezone.utc).isoformat()

            try:
                try:
                    bars = self.calls.get_bars(symbol, time_interval, epoch_start_str, epoch_stop_str,
                                               adjustment='raw').df
                except TypeError:
                    # If you query a timeframe with no data the API throws a Nonetype issue so just return something
                    #  empty if that happens
                    return pd.DataFrame(columns=['time', 'open', 'high', 'low', 'close', 'volume'])
            except AlpacaAPIError as e:
                if e.code == 42210000:
                    warning_string = "Your alpaca subscription does not permit querying data from the last 15 " \
                                     "minutes. Blankly is adjusting your query."
                    utils.info_print(warning_string)
                    epoch_stop = time.time() - (build_minute() * 15)
                    if epoch_stop >= epoch_start:
                        try:
                            return self.get_product_history(symbol, epoch_start, epoch_stop, resolution)
                        except TypeError:
                            # If you query a timeframe with no data the API throws a Nonetype issue so just
                            #  return something
                            #  empty if that happens
                            return pd.DataFrame(columns=['time', 'open', 'high', 'low', 'close', 'volume'])
                    else:
                        warning_string = "No data range queried after time adjustment."
                        utils.info_print(warning_string)
                        return pd.DataFrame(columns=['time', 'open', 'high', 'low', 'close', 'volume'])
                else:
                    raise e
            bars.rename(columns={"t": "time", "o": "open", "h": "high", "l": "low", "c": "close", "v": "volume"},
                        inplace=True)

            if bars.empty:
                return pd.DataFrame(columns=['time', 'open', 'high', 'low', 'close', 'volume'])
            return utils.get_ohlcv(bars, row_divisor, from_zero=False)
        else:
            # This runs yfinance on the symbol
            return self.parse_yfinance(symbol, epoch_start, epoch_stop, resolution)

    def overridden_history(self, symbol, epoch_start, epoch_stop, resolution_seconds, **kwargs) -> pd.DataFrame:
        to = kwargs['to']
        # If it's a string alpaca is an edge case where epoch can be used
        if to is not None and isinstance(to, str):
            to = None
        if to:
            if not self.user_preferences['settings']['alpaca']['use_yfinance']:
                resolution_seconds = self.valid_resolutions[min(range(len(self.valid_resolutions)),
                                                                key=lambda j: abs(self.valid_resolutions[j] -
                                                                                  resolution_seconds))]
                resolution_lookup = {
                    60: '1Min',
                    300: '5Min',
                    900: '15Min',
                    86400: '1Day'
                }

                time_interval = resolution_lookup[resolution_seconds]

                def find_last_n_points(epoch_start_, epoch_stop_):
                    frames = []

                    # This just overestimates and then chops off the remainder
                    while epoch_start_ <= epoch_stop_:
                        # Create an end time by moving after the start time by 1000 datapoints
                        frames.append(self.calls.get_bars(symbol, time_interval, limit=10000,
                                                          start=utils.iso8601_from_epoch(epoch_start_)).df)
                        epoch_start_ += resolution_seconds * 10000

                    for i in range(len(frames)):
                        series = []
                        for j in frames[i].index:
                            series.append(j.timestamp())
                        frames[i]['time'] = pd.Series(series).values
                        frames[i] = frames[i].reset_index(drop=True)

                    if len(frames) != 0:
                        response_ = pd.concat(frames, ignore_index=True)
                    else:
                        response_ = pd.DataFrame(columns=['time', 'open', 'high', 'low', 'close', 'volume'])

                    return response_

                # Create an empty dataframe here
                response = pd.DataFrame(columns=['time', 'open', 'high', 'low', 'close', 'volume'])
                batched_ranges = [epoch_start, epoch_stop]
                # Batch it until we have enough points
                tries = 2
                while len(response) < to:
                    response = pd.concat([response, find_last_n_points(batched_ranges[0], batched_ranges[1])])
                    batched_ranges[0] -= resolution_seconds * tries
                    batched_ranges[1] -= resolution_seconds * tries
                    tries = tries * 2

                response = response[['time', 'open', 'high', 'low', 'close', 'volume']]

                response = response.sort_values(by=['time'], ignore_index=True)

                response = response.tail(to)

                response = response.astype({
                    'time': int,
                    'open': float,
                    'high': float,
                    'low': float,
                    'close': float,
                    'volume': float,
                })
            else:
                # Overestimate the difference for yfinance
                epoch_start = epoch_stop - (epoch_stop - epoch_start) * 1.5

                return self.parse_yfinance(symbol, epoch_start, epoch_stop, resolution_seconds).tail(to).reset_index()
        else:
            response = self.get_product_history(symbol,
                                                epoch_start,
                                                epoch_stop,
                                                int(resolution_seconds))

        return response

    @staticmethod
    def __convert_times(date):  # There aren't any usages of this
        # convert start_date to datetime object
        if isinstance(date, str):
            date = dateparser.parse(date)
        elif isinstance(date, float):
            date = dt.fromtimestamp(date)

        # end_date object is naive datetime, so need to convert
        if date.tzinfo is None or date.tzinfo.utcoffset(date) is None:
            date = date.replace(tzinfo=timezone.utc)

        return date

    def get_order_filter(self, symbol: str):
        assert isinstance(self.calls, alpaca_trade_api.REST)
        current_price = self.get_price(symbol)

        products = self.get_products()

        product = None
        for i in products:
            if i['symbol'] == symbol:
                product = i
                break
        if product is None:
            raise APIException("Symbol not found.")

        exchange_specific = product['exchange_specific']
        fractionable = exchange_specific['fractionable']

        if fractionable:
            quote_increment = 1e-9
            min_funds_buy = 1
            min_funds_sell = 1e-9 * current_price

            # base_min_size = product['base_min_size']
            base_max_size = product['base_max_size']
            # base_increment = product['base_increment']
            min_price = 0.0001
            max_price = 10000000000

            # Guaranteed nano share if fractionable
            base_min_size = 1e-9
            base_increment = 1e-9
        else:
            quote_increment = current_price
            min_funds_buy = current_price
            min_funds_sell = current_price

            # base_min_size = product['base_min_size']
            base_max_size = product['base_max_size']
            # base_increment = product['base_increment']
            min_price = 0.0001
            max_price = 10000000000

            # Always 1 if not fractionable
            base_min_size = 1
            base_increment = 1

        max_funds = current_price * 10000000000

        return {
            "symbol": symbol,
            "base_asset": symbol,
            "quote_asset": 'USD',
            "max_orders": 500,  # More than this and we can't calculate account value (alpaca is very bad)
            "limit_order": {
                "base_min_size": 1,  # Minimum size to buy
                "base_max_size": base_max_size,  # Maximum size to buy
                "base_increment": 1,  # Specifies the minimum increment for the base_asset.

                "price_increment": min_price,  # TODO test this at market open

                "min_price": min_price,
                "max_price": max_price,
            },
            'market_order': {
                "fractionable": fractionable,

                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": base_max_size,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment for the base_asset.

                "quote_increment": quote_increment,  # Specifies the min order price as well as the price increment.
                "buy": {
                    "min_funds": min_funds_buy,
                    "max_funds": max_funds,
                },
                "sell": {
                    "min_funds": min_funds_sell,
                    "max_funds": max_funds,
                },
            },
            "exchange_specific": {
                "id": exchange_specific['id'],
                "class": exchange_specific['class'],
                "exchange": exchange_specific['exchange'],
                "status": exchange_specific['status'],
                "tradable": exchange_specific['tradable'],
                "marginable": exchange_specific['marginable'],
                "shortable": exchange_specific['shortable'],
                "easy_to_borrow": exchange_specific['easy_to_borrow'],
                "price": current_price
            }
        }

    def get_price(self, symbol) -> float:
        assert isinstance(self.calls, alpaca_trade_api.REST)
        response = self.calls.get_latest_trade(symbol=symbol)
        return float(response['p'])

    @staticmethod
    def parse_yfinance(symbol: str, epoch_start: [int, float], epoch_stop: [int, float], resolution: int):
        try:
            import yfinance

            start_date = dt.fromtimestamp(epoch_start, tz=timezone.utc)
            stop_date = dt.fromtimestamp(epoch_stop, tz=timezone.utc)
            ticker = yfinance.Ticker(symbol)
            result = ticker.history(start=start_date, end=stop_date, interval=number_interval_to_string(resolution))

            result['time'] = result.index.astype(int) // 10 ** 9
            result = result[['Open', 'High', 'Low', 'Close', 'Volume', 'time']].reset_index()

            result = result.rename(columns={
                'time': 'time',
                'Open': 'open',
                'High': 'high',
                'Low': 'low',
                'Close': 'close',
                'Volume': 'volume'
            })

            result = result[['time', 'open', 'high', 'low', 'close', 'volume']]

            return result
        except ImportError:
            raise ImportError("To use yfinance to download data please pip install yfinance")



"""
    Keyless interface definition
    Copyright (C) 2022 Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pandas as pd


from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.utils.utils import AttributeDict, get_base_asset, get_quote_asset, aggregate_prices_by_resolution, \
    extract_price_by_resolution


# This just happens to also inherit from the exchange interface
class KeylessAPI(ExchangeInterface):
    def __init__(self, maker_fee, taker_fee, price_readers):
        if not isinstance(price_readers, list):
            price_readers = [price_readers]

        self.__final_prices = {}
        prices: dict = {}

        for price_reader in price_readers:
            data = price_reader.data
            for symbol in data:
                symbol_info = price_reader.prices_info[symbol]
                resolution = symbol_info['resolution']

                # Add each symbol to the final prices without doing any processing
                if symbol in prices:
                    prices[symbol] = pd.concat([prices[symbol], data[symbol]])
                else:
                    prices[symbol] = data[symbol]

                self.__final_prices = aggregate_prices_by_resolution(self.__final_prices, symbol, resolution,
                                                                     data[symbol])

        self.__products = None
        self.__accounts = None

        self.fees = {
            'maker_fee_rate': maker_fee,
            'taker_fee_rate': taker_fee
        }

        super().__init__('keyless', self)

    def init_exchange(self):
        pass

    def __invalid_live(self):
        raise RuntimeError("Cannot use a keyless exchange in live trading. Please insert keys and begin using exchanges"
                           " such as blankly.Alpaca() or blankly.CoinbasePro().")

    def get_exchange_type(self):
        return 'keyless'

    def get_account(self, symbol: str = None) -> AttributeDict:
        return AttributeDict(self.__accounts)

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        return extract_price_by_resolution(self.__final_prices, symbol, epoch_start, epoch_stop, resolution)

    def get_products(self):
        self.__invalid_live()

    def market_order(self, symbol: str, side: str, size: float):
        self.__invalid_live()

    def limit_order(self, symbol: str, side: str, price: float, size: float):
        self.__invalid_live()

    def cancel_order(self, symbol: str, order_id: str):
        self.__invalid_live()

    def get_open_orders(self, symbol: str = None):
        self.__invalid_live()

    def get_order(self, symbol: str, order_id: str):
        self.__invalid_live()

    def get_fees(self, symbol) -> dict:
        return self.fees

    def get_order_filter(self, symbol: str) -> dict:
        return {
            "symbol": symbol,
            "base_asset": get_base_asset(symbol),
            "quote_asset": get_quote_asset(symbol),
            "max_orders": 1000000000000000,
            "limit_order": {
                "base_min_size": 0.000000001,  # Minimum size to buy
                "base_max_size": 1000000000000000,  # Maximum size to buy
                "base_increment": 0.000000001,  # Specifies the minimum increment
                # for the base_asset.
                "price_increment": 0.000000001,

                "min_price": 0.000000001,
                "max_price": 1000000000000000,
            },
            'market_order': {
                "fractionable": True,

                "base_min_size": 0.000000001,  # Minimum size to buy
                "base_max_size": 1000000000000000,  # Maximum size to buy
                "base_increment": 0.000000001,  # Specifies the minimum increment

                "quote_increment": 0.000000001,  # Specifies the min order price as well
                # as the price increment.
                "buy": {
                    "min_funds": 0.000000001,
                    "max_funds": 1000000000000000,
                },
                "sell": {
                    "min_funds": 0.000000001,
                    "max_funds": 1000000000000000,
                },
            },
            "exchange_specific": {}
        }

    def get_price(self, symbol: str, time=None):
        self.__invalid_live()



import time

from blankly.utils.utils import isolate_specific


def switch_type(stream):
    if stream == "ticker":
        return trade, \
               trade_interface, \
               "time, sequence, price, size, bestAsk, bestAskSize, bestBid, bestBidSize\n"
    elif stream == "level2":
        return no_callback, \
               no_callback, \
               ""
    else:
        return no_callback, no_callback, ""


def no_callback(message):
    return message


def trade(message):
    line = str(time.time()) + "," + message["sequence"] + "," + message["price"] + "," + \
           message["size"] + "," + message["bestAsk"] + "," + message["bestAskSize"] + "," + \
           message["bestBid"] + "," + message["bestBidSize"] + "\n"
    return line


def trade_interface(message):
    # renames = [
    #     ["sequence", "trade_id"]
    # ]
    #
    # message = rename_to(renames, message)

    needed = [
        ["symbol", str],
        ["price", float],
        ["time", float],
        ["trade_id", int],
        ["size", float]
    ]
    message['symbol'] = message['topic'].split(":", 1)[1]
    message['trade_id'] = message['data'].pop('sequence')
    message['price'] = message['data'].pop('price')
    message['size'] = message['data'].pop('size')

    message["time"] = time.time()

    isolated = isolate_specific(needed, message)

    return isolated



"""
    Base Exchange object.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import abc


class ABCBaseExchange(abc.ABC):

    @abc.abstractmethod
    def get_interface(self):
        pass

    @abc.abstractmethod
    def get_name(self) -> str:
        pass

    @abc.abstractmethod
    def get_type(self) -> str:
        pass



import time
import pandas as pd
import blankly.utils.time_builder
import blankly.utils.utils as utils
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.interfaces.okx.okx_api import MarketAPI, AccountAPI, TradeAPI, ConvertAPI, FundingAPI, PublicAPI
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.utils.exceptions import APIException, InvalidOrder


class OkxInterface(ExchangeInterface):
    def __init__(self, exchange_name, authenticated_api):
        self._market: MarketAPI = authenticated_api['market']
        self._account: AccountAPI = authenticated_api['account']
        self._trade: TradeAPI = authenticated_api['trade']
        self._convert: ConvertAPI = authenticated_api['convert']
        self._funding: FundingAPI = authenticated_api['funding']
        self._public: PublicAPI = authenticated_api['public']
        super().__init__(exchange_name, authenticated_api, valid_resolutions=[60, 180, 300, 1800, 3600, 7200,
                                                                              14400, 21600, 43200, 86400, 604800,
                                                                              2629746, 7889238, 15778476, 31556952])

    def init_exchange(self):
        # This is purely an authentication check which can be disabled in settings
        # todo: change get_fee_rates
        fees = self._account.get_fee_rates(instType='SPOT', instId='BTC-USDT')
        try:
            if fees['msg'] == "Invalid API Key":
                raise LookupError("Invalid API Key. Please check if the keys were input correctly into your "
                                  "keys.json.")
        except KeyError:
            pass

    def get_products(self):
        instrument_type = "SPOT"
        needed = self.needed['get_products']
        products = self._public.get_instruments(instrument_type)

        for i in range(len(products['data'])):
            products['data'][i]["symbol"] = products['data'][i]["instId"]
            products['data'][i]["base_asset"] = products['data'][i]["baseCcy"]
            products['data'][i]["quote_asset"] = products['data'][i]["quoteCcy"]
            products['data'][i]["base_min_size"] = products['data'][i]["minSz"]
            products['data'][i]["base_max_size"] = 1000000000
            products['data'][i]["base_increment"] = products['data'][i]["tickSz"]
            products['data'][i] = utils.isolate_specific(needed, products['data'][i])
        return products['data']


    def get_account(self, symbol=None) -> utils.AttributeDict:
        """
           Get all currencies in an account, or sort by symbol
           Args:
               symbol (Optional): Filter by particular symbol
               These arguments are mutually exclusive
        """

        symbol = super().get_account(symbol=symbol)
        needed = self.needed['get_account']
        accounts = self._account.get_account()

        parsed_dictionary = utils.AttributeDict({})
        #We have to sort through it if the accounts are none
        if symbol is not None:
            accounts_specific = self._account.get_account(symbol)
            if accounts_specific['code'] == 51000 or accounts_specific['code'] == 51001:
                raise ValueError("Symbol not found")
            for i in accounts_specific['data']:
                parsed_value = utils.isolate_specific(needed, i)
                dictionary = utils.AttributeDict({
                    'available': float(parsed_value['exchange_specific']['details'][0]['availBal']), # accounts_specific['data'][0]['details'][0]['availBal'],
                    'hold': float(parsed_value['exchange_specific']['details'][0]['frozenBal'])
                })
                return dictionary
        for i in range(len(accounts['data'][0]['details'])):
            parsed_dictionary[accounts['data'][0]['details'][i]['ccy']] = utils.AttributeDict({
                'available': float(accounts['data'][0]['details'][i]['availBal']),
                'hold': float(accounts['data'][0]['details'][i]['frozenBal'])
            })

        return parsed_dictionary

    @utils.order_protection
    def market_order(self, symbol, side, size) -> MarketOrder:
        """
            Used for buying or selling market orders
            Args:
                symbol: currency to buy
                side: buy/sell
                size: desired amount of base currency to buy or sell
        """
        needed = self.needed['market_order']
        """
            {
                "result": "true",
                "request_id": "",
                "client_oid": "b1234",
                "order_id": "131801215763791872",
                "error_message": "",
                "error_code": "0"
            }
        """
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))
        order = {
            'symbol': symbol,
            'size': size,
            'side': side,
            'type': 'market'
        }

        response = self._trade.place_order(symbol, 'cash', side, 'market', size)
        if len(response['data'][0]['sMsg']) != 0:
            raise InvalidOrder("Invalid Order: " + response['data'][0]["sMsg"])
        response["created_at"] = time.time()
        response["id"] = response['data'][0]['ordId']
        response["status"] = response['data'][0]['sCode']
        response["symbol"] = symbol
        response["size"] = size
        response["side"] = side
        response["type"] = 'market'
        response = utils.isolate_specific(needed, response)
        return MarketOrder(order, response, self)

    @utils.order_protection
    def limit_order(self, symbol, side, price, size) -> LimitOrder:
        """
           Used for buying or selling limit orders
           Args:
               symbol: currency to buy
               side: buy/sell
               price: price to set limit order
               size: amount of currency (like BTC) for the limit to be valued
        """
        needed = self.needed['limit_order']
        if self.should_auto_trunc:
            result = utils.trunc(size, self.get_asset_precision(symbol))
            if result != 0:
                size = result
        order = {
            'symbol': symbol,
            'side': side,
            'size': size,
            'price': price,
            'type': 'limit',
        }

        response = self._trade.place_order(symbol, 'cash', side, 'limit', size, px=price)
        if len(response['data'][0]['sMsg']) != 0:
            raise InvalidOrder("Invalid Order: " + response['data'][0]["sMsg"])

        response_details = self._trade.get_orders(symbol, ordId=response['data'][0]['ordId'])

        response_details["created_at"] = response_details['data'][0]['cTime']
        response_details["id"] = response_details['data'][0]['ordId']
        response_details["status"] = response_details['data'][0]['state']
        response_details["symbol"] = response_details['data'][0]['instId']
        response_details["size"] = response_details['data'][0]['sz']
        response_details["side"] = response_details['data'][0]['side']
        response_details["price"] = response_details['data'][0]['px']
        response_details["type"] = response_details['data'][0]['ordType']
        response_details["time_in_force"] = 'GTC'

        response_details = utils.isolate_specific(needed, response_details)
        return LimitOrder(order, response_details, self)

    def cancel_order(self, symbol: str, order_id: str) -> dict:
        """
        Returns:
            dict: Containing the order_id of cancelled order. Example::
            { "client_oid": "order123", }
        """
        response = self._trade.cancel_order(symbol, ordId=order_id)
        return {"order_id": response['data'][0]['ordId']}

    def get_open_orders(self,
                        symbol: str = None) -> list:
        """
        List open orders
        """
        """
        [
            {
                "client_oid":"oktspot86",
                "created_at":"2019-03-20T03:28:14.000Z",
                "filled_notional":"0",
                "filled_size":"0",
                "funds":"",
                "instrument_id":"BTC-USDT",
                "notional":"",
                "order_id":"2511109744100352",
                "order_type":"0",
                "price":"3594.7",
                "price_avg":"",
                "product_id":"BTC-USDT",
                "side":"buy",
                "size":"0.001",
                "status":"open",
                "fee_currency":"BTC",
                "fee":"-0.01",
                "rebate_currency":"open",
                "rebate":"0.05",
                "state":"0",
                "timestamp":"2019-03-20T03:28:14.000Z",
                "type":"limit"
            }
        ]
        """
        if symbol is None:
            orders = self._trade.get_order_list()
        else:
            orders = self._trade.get_order_list(instId=symbol)

        if len(orders['data']) == 0:
            return []
        if orders['code'].startswith('5'):
            raise InvalidOrder("Invalid Order: " + str(orders))

        for i in range(len(orders['data'])):
            orders['data'][i]["created_at"] = orders['data'][i]["cTime"]
            needed = self.choose_order_specificity(orders['data'][i]['ordType'])
            orders['data'][i]["symbol"] = orders['data'][i]['instId']
            orders['data'][i]["id"] = orders['data'][i]['ordId']
            orders['data'][i]["status"] = orders['data'][i]['state']
            orders['data'][i]["size"] = orders['data'][i]['sz']
            orders['data'][i]["price"] = orders['data'][i]['px']
            orders['data'][i]["type"] = orders['data'][i]["ordType"]
            if orders['data'][i]["ordType"] == "limit":
                orders['data'][i]["time_in_force"] = 'GTC'
            orders['data'][i] = utils.isolate_specific(needed, orders['data'][i])

        return orders['data']

    def get_order(self, symbol, order_id) -> dict:
        response = self._trade.get_orders(symbol, ordId=order_id)

        if 'message' in response:
            raise APIException("Invalid: " + str(response['msg']) + ", was the order canceled?")

        if response['data'][0]['ordType'] == 'market':
            needed = self.needed['market_order']
        elif response['data'][0]['ordType'] == 'limit':
            needed = self.needed['limit_order']
            response['price'] = response['data'][0]['px']
        else:
            needed = self.needed['market_order']

        response["symbol"] = response['data'][0]['instId']
        response["created_at"] = response['data'][0]['cTime']
        response["size"] = response['data'][0]['sz']
        response["side"] = response['data'][0]['side']
        response["type"] = response['data'][0]['ordType']
        response["id"] = response['data'][0]['ordId']
        response["status"] = response['data'][0]['state']
        response["time_in_force"] = 'GTC'
        return utils.isolate_specific(needed, response)

    def get_fees(self, symbol) -> dict:
        needed = self.needed['get_fees']
        """
        {
            "category": "1",
            "maker": "0.0002",
            "taker": "0.0005",
            "timestamp": "2019-12-11T11:02:31.360Z"
        }
        """
        if symbol is None:
            raise ValueError("There was no symbol inputted, please try again.")
        else:
            fees = self._account.get_fee_rates(instType="SPOT", instId=symbol)
        fees['taker_fee_rate'] = fees['data'][0]['taker']
        fees['maker_fee_rate'] = fees['data'][0]['maker']
        return utils.isolate_specific(needed, fees)

    def overridden_history(self, symbol, epoch_start, epoch_stop, resolution, **kwargs) -> pd.DataFrame:
        """
        Okx is strange
        """
        to = kwargs['to']

        if isinstance(to, str):
            epoch_start = epoch_start - resolution
        elif isinstance(to, int):
            epoch_start = epoch_start - resolution

        return self.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        if epoch_stop > time.time():
            epoch_stop = int(time.time())

        resolution = blankly.time_builder.time_interval_to_seconds(resolution)

        init_epoch_start = int(epoch_start * 1000)
        init_epoch_stop = int(epoch_stop * 1000)

        accepted_grans = [60, 180, 300, 1800, 3600, 7200, 14400, 21600, 43200, 86400, 604800, 2629746, 7889238, 15778476, 31556952]

        if resolution not in accepted_grans:
            utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
            resolution = accepted_grans[min(range(len(accepted_grans)),
                                            key=lambda i: abs(accepted_grans[i] - resolution))]

        resolution = int(resolution)

        lookup_dict = {
            60: '1m',
            180: '3m',
            300: '5m',
            900: '15m',
            1800: '30m',
            3600: '1H',
            7200: '2H',
            14400: '4H',
            21600: '6H',
            43200: '12H',
            86400: '1D',
            604800: '1W',
            2629746: '1M',
            7889238: '3M',
            15778476: '6M',
            31556952: '1Y'
        }

        gran_string = lookup_dict[resolution]

        history = []
        while init_epoch_start < init_epoch_stop + resolution * 100 * 1100:
            response = self._market.get_history_candlesticks(symbol, after=init_epoch_start, bar=gran_string, limit=100)

            if len(response['data']) != 0:
                history = history + response['data']
            init_epoch_start = init_epoch_start + (100 * 1000 * resolution)

        history.sort(key=lambda x: x[0])
        new_history = list(set(tuple(sub) for sub in history))
        new_history.sort(key=lambda x: x[0])
        df = pd.DataFrame(new_history, columns=['time', 'open', 'high', 'low', 'close', 'volume', 'volume_currency'])
        df = df.drop(columns=['volume_currency'])
        df[['time']] = df[['time']].astype('int64').div(1000).astype(int)
        df[['open', 'high', 'low', 'close', 'volume']] = df[['open', 'high', 'low', 'close', 'volume']].astype(float)
        # df.plot(x='time', y='close', kind='scatter')
        df = df[df['time'] >= epoch_start]
        df = df[df['time'] <= epoch_stop]
        # plt.show()

        return df

    def get_order_filter(self, symbol: str) -> dict:
        instrument_type = "SPOT"
        response = self._public.get_instruments(instrument_type)
        products = None
        for i in response["data"]:
            if i["instId"] == symbol:
                products = i
                break

        base_min_size = float(products.pop('minSz'))
        base_increment = float(products.pop('tickSz'))

        return {
            "symbol": products.pop('uly'),
            "base_asset": products.pop('baseCcy'),
            "quote_asset": products.pop('quoteCcy'),
            "max_orders": 1000000000000,
            "limit_order": {
                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": 100000000,  # Maximum size to buy, this value was hardcoded
                "base_increment": base_increment,  # Specifies the minimum increment for the base_asset.
                "price_increment": 0.1,

                "min_price": 0.1,
                "max_price": 9999999999,
            },
            'market_order': {
                "fractionable": True,

                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": 100000000,  # Maximum size to buy, this value was hardcoded
                "base_increment": base_increment,  # Specifies the minimum increment

                "quote_increment": 0.1,  # Specifies the min order price as well as the price increment.
                "buy": {
                    "min_funds": 0.1,  # This value was hardcoded
                    "max_funds": 1000000,  # This value was hardcoded
                },
                "sell": {
                    "min_funds": 0.1,  # This value was hardcoded
                    "max_funds": 1000000,  # This value was hardcoded
                },
            },
            "exchange_specific": {**products}
        }

    def get_price(self, symbol) -> float:
        """
        Returns just the price of a currency pair.
        """
        response = self._market.get_index_ticker(instId=symbol)
        if len(response['msg']) != 0:
            raise APIException("Error: " + response['msg'])
        return float(response['data'][0]['idxPx'])





"""
    Keyless exchange definition & setup
    Copyright (C) 2022 Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import blankly
from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.exchanges.interfaces.keyless.keyless_api import KeylessAPI
from blankly.data.data_reader import PriceReader


class KeylessExchange(Exchange):
    def __init__(self, maker_fee=0, taker_fee=0, portfolio_name=None, settings_path=None,
                 price_reader: [PriceReader, list] = None):
        Exchange.__init__(self, "keyless", portfolio_name, settings_path)

        self.calls = KeylessAPI(maker_fee, taker_fee, price_reader)

        self.interface = PaperTradeInterface(self.calls)

        # This one must be exported manually
        blankly.reporter.export_used_exchange("keyless")

    """
    Builds information about the asset on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        This determines the internal properties of the exchange block.
        Should be implemented per-class because it requires different types of interaction with each exchange.
        """
        # TODO Populate this with useful information
        return self.interface.get_account(symbol)

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_products()

    def get_direct_calls(self) -> ABCExchangeInterface:
        return self.calls



"""
    Abstraction for creating interval driven user strategies
    Copyright (C) 2021  Emerson Dove, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import threading
import typing
import enum


import blankly
from blankly.exchanges.abc_base_exchange import ABCBaseExchange
from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.frameworks.strategy.strategy_state import StrategyState
from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import AttributeDict


class EventType(enum.Enum):
    price_event = 'price_event'
    bar_event = 'bar_event'
    scheduled_event = 'scheduled_event'
    arbitrage_event = 'arbitrage_event'
    orderbook_event = "orderbook_event"


class StrategyBase:
    __exchange: ABCBaseExchange
    interface: ABCBaseExchangeInterface

    def __init__(self, exchange: ABCBaseExchange, interface: ABCBaseExchangeInterface, model):
        """
        Create a new strategy object. A strategy can be used to run your code live while be backtestable and modular
         across exchanges.
         Args:
             exchange: An exchange object. This can be created by doing something similar to exchange = blankly.Alpaca()

        Function Signatures:
        init(symbol: str, state: blankly.StrategyState)
        price_event(price: float, symbol: str, state: blankly.StrategyState)
        orderbook_event(tick: dict, symbol: str, state: blankly.StrategyState)
        bar_event(bar: dict, symbol: str, state: blankly.StrategyState)
        teardown(blankly.StrategyState)
        """
        self.remote_backtesting = blankly._backtesting
        self.__exchange = exchange
        self.interface = interface

        self.ticker_manager = blankly.TickerManager(self.__exchange.get_type(), '')
        self.orderbook_manager = blankly.OrderbookManager(self.__exchange.get_type(), '')

        # Attempt to report the strategy
        blankly.reporter.export_strategy(self)

        # Create a lock for the teardown so nothing happens while it's going on
        self.lock = threading.Lock()

        # This will be updated when the teardown() function completes
        self.torndown = False

        self.schedulers = []

        self.interface = interface

        self.orderbook_websockets = []
        self.ticker_websockets = []
        self.model = model

    def add_price_event(self, callback: typing.Callable, symbol: str, resolution: typing.Union[str, float],
                        init: typing.Callable = None, teardown: typing.Callable = None, synced: bool = False,
                        variables: dict = None):
        """
        Add Price Event. This will provide you with an updated price every time the callback is run
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the price event for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            synced: Sync the function to
            variables: A dictionary to initialize the state's internal values
        """
        self.__custom_price_event(callback=callback, symbol=symbol, resolution=resolution, init=init, synced=synced,
                                  teardown=teardown, variables=variables, type_=EventType.price_event)

    def add_scheduled_event(self, callback: typing.Callable, resolution: typing.Union[str, float],
                            init: typing.Callable = None, teardown: typing.Callable = None,
                            synced: bool = False, variables: dict = None):
        """
        Add a scheduled event. This will call the callback at the rate defined in the resolution

        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            synced: Sync the function to
            variables: A dictionary to initialize the state's internal values
        """
        self.__custom_price_event(callback=callback, symbol=None, resolution=resolution, init=init, synced=synced,
                                  teardown=teardown, variables=variables, type_=EventType.scheduled_event)

    def add_arbitrage_event(self, callback: typing.Callable, symbols: list, resolution: typing.Union[str, float],
                            init: typing.Callable = None, teardown: typing.Callable = None, synced: bool = False,
                            variables: dict = None):
        """
        Add Price Event. This will provide you with an updated price every time the callback is run
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbols: A list of symbols to get data for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            synced: Sync the function to
            variables: A dictionary to initialize the state's internal values
        """
        self.__custom_price_event(callback=callback, symbol=symbols, resolution=resolution, init=init, synced=synced,
                                  teardown=teardown, variables=variables, type_=EventType.arbitrage_event)

    def add_bar_event(self, callback: typing.Callable, symbol: str, resolution: typing.Union[str, float],
                      init: typing.Callable = None, teardown: typing.Callable = None, variables: dict = None):
        """
        The bar event sends a dictionary of {open, high, low, close, volume} which has occurred in the interval.
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the price event for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            variables: A dictionary to initialize the state's internal values
        """
        self.__custom_price_event(type_=EventType.bar_event, synced=True, callback=callback, symbol=symbol,
                                  resolution=resolution, init=init, teardown=teardown, variables=variables)

    def __custom_price_event(self,
                             type_: EventType,
                             callback: typing.Callable = None,
                             symbol: [str, list] = None,
                             resolution: typing.Union[str, float] = None,
                             init: typing.Callable = None,
                             synced: bool = False,
                             teardown: typing.Callable = None, variables: dict = None):
        """
        Add Price Event
        Args:
            type_: The type of event that the price event refers to
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the price event for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            synced: Sync the function to
            variables: Initial dictionary to write into the state variable
        """
        # Make sure variables is always an empty dictionary if None
        if variables is None:
            variables = {}
        resolution = time_interval_to_seconds(resolution)

        variables_ = AttributeDict(variables)
        state = StrategyState(self, variables_, symbol, resolution=resolution)

        # Use the API
        self.schedulers.append(
            blankly.Scheduler(self.model.rest_event, resolution,
                              initially_stopped=True,
                              synced=synced,
                              callback=callback,
                              resolution=resolution,
                              variables=variables_,
                              state=state,
                              type=type_,
                              init=init,
                              teardown=teardown,
                              symbol=symbol)
        )

        # Export a new symbol to the backend
        if symbol is not None:
            if isinstance(symbol, list):
                for i in symbol:
                    blankly.reporter.export_used_symbol(i)
            else:
                blankly.reporter.export_used_symbol(symbol)

    @staticmethod
    def __websocket_callback(tick, symbol, user_callback, state_object):
        user_callback(tick, symbol, state_object)

    def add_tick_event(self, callback: callable, symbol: str, init: callable = None, teardown: callable = None,
                       variables: dict = None):
        """
        Add a tick event - This will call the callback everytime the exchange provides a change in the price due to a
         trade occurring
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the orderbook for
            init: Callback function to allow a setup for the strategy variable. This
                can be used for accumulating price data
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful:
            variables: A dictionary to initialize the state's internal values
        """
        if variables is None:
            variables = {}

        state = StrategyState(self, AttributeDict(variables), symbol=symbol)

        self.ticker_manager.create_ticker(self.__websocket_callback, initially_stopped=True,
                                          override_symbol=symbol,
                                          symbol=symbol,
                                          user_callback=callback,
                                          variables=variables,
                                          state=state)

        self.ticker_websockets.append([symbol, self.__exchange.get_type(), init, state, teardown])

    def add_orderbook_event(self, callback: callable, symbol: str, init: typing.Callable = None,
                            teardown: typing.Callable = None, variables: dict = None):
        """
        Add Orderbook Event - This will call the given callback everytime the exchange provides a change in the
         orderbook
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the orderbook for
            init: Callback function to allow a setup for the strategy variable. This
                can be used for accumulating price data
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful:
            variables: A dictionary to initialize the state's internal values
        """
        # Make sure variables is always an empty dictionary if None
        if variables is None:
            variables = {}

        state = StrategyState(self, AttributeDict(variables), symbol=symbol)

        # since it's less than 10 sec, we will just use the websocket feed - exchanges don't like fast calls
        self.orderbook_manager.create_orderbook(self.__websocket_callback, initially_stopped=True,
                                                override_symbol=symbol,
                                                symbol=symbol,
                                                user_callback=callback,
                                                variables=variables,
                                                state=state)

        self.ticker_websockets.append([symbol, self.__exchange.get_type(), init, state, teardown])

    def start(self):
        """
        Run your model live!

        Simply call this function to take your strategy configuration live on your exchange
        """
        raise NotImplementedError

    def teardown(self):
        raise NotImplementedError

    def time(self) -> float:
        raise NotImplementedError

    def backtest(self,
                 to: str = None,
                 initial_values: dict = None,
                 start_date: typing.Union[str, float, int] = None,
                 end_date: typing.Union[str, float, int] = None,
                 settings_path: str = None,
                 **kwargs
                 ) -> BacktestResult:
        raise NotImplementedError



import prompt_toolkit
from functools import lru_cache
from prompt_toolkit import print_formatted_text, formatted_text
from prompt_toolkit.formatted_text import to_formatted_text

import questionary.constants
from yaspin import yaspin, Spinner as YaspinSpinner
from yaspin.core import Yaspin
from yaspin.spinners import Spinners

BOLD = '\033[1m'

STYLE = questionary.Style(
    [
        ('qmark', 'fg:#5f819d'),  # token in front of the question
        ('work', 'fg:#5f819d'),  # token in front of work message
        ('failure', 'fg:#ff726f'),  # token in front of failure message
        ('success', 'fg:#4BCA81'),  # token in front of success message
        ('question', 'bold'),  # question text
        ('answer', 'fg:#FF9D00 bold'),  # submitted answer text behind the question
        ('pointer', ''),  # pointer used in select and checkbox prompts
        ('selected', ''),  # style for a selected item of a checkbox
        ('separator', ''),  # separator in lists
        ('instruction', ''),  # user instructions for select, rawselect, checkbox
        ('text', ''),  # any other text
        ('instruction', ''),  # user instructions for select, rawselect, checkbox
    ]
)
questionary.constants.DEFAULT_STYLE = STYLE

from questionary import confirm, text, select


def fprint(text, **kwargs):
    if isinstance(text, str):
        text = [('class:text', text)]
    print_formatted_text(to_formatted_text(text), style=STYLE, **kwargs)


def print_work(text):
    fprint([('class:work', '*'), ('class:question', ' ' + text)])


def print_failure(text):
    fprint([('class:failure', '✘ '), ('class:question', text)])


def print_success(text):
    fprint([('class:success', '✔ '), ('class:question', text)])


@lru_cache(None)
def text_spinner(
        text='BLANKLY',
        left='[',
        right=']',
        interval=80
):
    fullstr = text + ' ' * (len(text)//2) + text[:-1]
    win_size = len(text)
    return YaspinSpinner([
        left + fullstr[i:i + win_size] + right
        for i in range(len(fullstr) - win_size, 0, -1)
    ], interval)


def show_spinner(text):
    return Spinner(text_spinner(), text=BOLD + text)


class Spinner(Yaspin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.outcome = None
        self.outcome_text = ''

    def fail(self, text):
        self.outcome = False
        self.outcome_text = text

    def ok(self, text):
        self.outcome = True
        self.outcome_text = text

    def stop(self):
        super().stop()
        if self.outcome is None:
            return

        if self.outcome:
            print_success(self.outcome_text)
        else:
            print_failure(self.outcome_text)



"""
    Datareader types & imports
    Copyright (C) 2022  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from enum import Enum as __Enum


class DataTypes(__Enum):
    ohlcv_csv = 0
    tick_csv = 1
    ohlcv_json = 2
    event_json = 3


from .data_reader import PriceReader, JsonEventReader, TickReader


"""
Some datatype examples


ohlcv_csv
symbol1, time, open, high, low, close, volume, symbol2, time, open, high, low, close, volume


ohlcv_json
{
    "symbol": {
        "open": [],
        "high": [],
        "low": [],
        "close": [],
        "volume": []
    }
}


event_json
{
    "event_type": {
        "time": [],
        "data": []
    }

}
"""



"""
    Interface for communicating with Oanda
    Copyright (C) 2021  Arun Annamalai

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import dateparser as dp
from datetime import datetime as dt
from typing import Union

import pandas as pd

from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.interfaces.oanda.oanda_api import OandaAPI
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.utils import utils as utils
from blankly.utils.exceptions import APIException, InvalidOrder


class OandaInterface(ExchangeInterface):
    def __init__(self, exchange_name: str, authenticated_api: OandaAPI):
        self.default_trunc = None
        super().__init__(exchange_name, authenticated_api, valid_resolutions=[5, 10, 15, 30, 60,
                                                                              60 * 2, 60 * 4, 60 * 5,
                                                                              60 * 10, 60 * 15, 60 * 30,
                                                                              60 * 60, 60 * 60 * 24,
                                                                              60 * 60 * 24 * 7,
                                                                              60 * 60 * 24 * 30])
        self.supported_multiples = {5: "S5", 10: "S10", 15: "S15", 30: "S30", 60: "M1", 60 * 2: "M2",
                                    60 * 4: "M4", 60 * 5: "M5", 60 * 10: "M10", 60 * 15: "M15", 60 * 30: "M30",
                                    60 * 60: "H1", 60 * 60 * 24: "D", 60 * 60 * 24 * 7: "W", 60 * 60 * 24 * 30: "M"}

        self.multiples_keys = list(self.supported_multiples.keys())

        self.unique_assets = None

        self.max_candles = 5000

        self.calls: OandaAPI

    def init_exchange(self):
        account_info = self.calls.get_account()
        if 'errorMessage' in account_info:
            raise LookupError(f"{account_info['errorMessage']}. \nTry toggling the \'sandbox\' setting "
                              "in your keys.json or check if the keys were input correctly into your "
                              "keys.json.")
        assert account_info['account']['id'] is not None, "Oanda exchange account does not exist"

        self.default_trunc = self._get_default_truncation()

        filtered_assets = []
        products = self.get_products()
        for i in products:
            base = utils.get_base_asset(i['symbol'])
            quote = utils.get_quote_asset(i['symbol'])
            if base not in filtered_assets:
                filtered_assets.append(base)
            if quote not in filtered_assets:
                filtered_assets.append(quote)

        self.unique_assets = filtered_assets

    def get_products(self) -> list:
        """
        Insturments response:
            {
                'name': 'GBP_NZD',
                'type': 'CURRENCY',
                'displayName': 'GBP/NZD',
                'pipLocation': -4,
                'displayPrecision': 5,
                'tradeUnitsPrecision': 0,
                'minimumTradeSize': '1',
                'maximumTrailingStopDistance': '1.00000',
                'minimumTrailingStopDistance': '0.00050',
                'maximumPositionSize': '0',
                'maximumOrderUnits': '100000000',
                'marginRate': '0.03',
                'guaranteedStopLossOrderMode': 'DISABLED',
                'tags': [{
                    'type': 'ASSET_CLASS',
                    'name': 'CURRENCY'
                }],
                'financing': {
                    'longRate': '-0.0153',
                    'shortRate': '-0.0093',
                    'financingDaysOfWeek': [{
                        'dayOfWeek': 'MONDAY',
                        'daysCharged': 1
                    }, {
                        'dayOfWeek': 'TUESDAY',
                        'daysCharged': 1
                    }, {
                        'dayOfWeek': 'WEDNESDAY',
                        'daysCharged': 1
                    }, {
                        'dayOfWeek': 'THURSDAY',
                        'daysCharged': 1
                    }, {
                        'dayOfWeek': 'FRIDAY',
                        'daysCharged': 1
                    }, {
                        'dayOfWeek': 'SATURDAY',
                        'daysCharged': 0
                    }, {
                        'dayOfWeek': 'SUNDAY',
                        'daysCharged': 0
                    }]
            }
        }
        """
        needed = self.needed['get_products']
        instruments = self.calls.get_account_instruments()['instruments']

        for instrument in instruments:
            instrument['symbol'] = self.__convert_symbol_to_blankly(instrument.pop('name'))
            instrument['base_asset'] = utils.get_base_asset(instrument['symbol'])
            instrument['quote_asset'] = utils.get_quote_asset(instrument['symbol'])
            instrument['base_min_size'] = float(instrument['minimumTradeSize'])
            instrument['base_max_size'] = float(instrument['maximumOrderUnits'])
            instrument['base_increment'] = 10 ** (-1 * int(instrument['tradeUnitsPrecision']))

        for i in range(len(instruments)):
            instruments[i] = utils.isolate_specific(needed, instruments[i])

        return instruments

    @utils.enforce_base_asset
    def get_account(self, symbol=None) -> utils.AttributeDict:
        if symbol is not None:
            symbol = self.__convert_blankly_to_oanda(symbol)
        positions_dict = utils.AttributeDict({})
        positions = self.calls.get_all_positions()['positions']
        for position in positions:
            # Split both the base and the quote
            # This the base split [0]
            positions_dict[position['instrument'].split('_')[0]] = utils.AttributeDict({
                'available': float(position['long']['units']) - float(position['short']['units']),
                'hold': 0.0
            })
            # This is the quote split [1]
            positions_dict[position['instrument'].split('_')[1]] = utils.AttributeDict({
                'available': float(position['long']['units']) - float(position['short']['units']),
                'hold': 0.0
            })

        def cash():
            account_dict = self.calls.get_account()['account']
            return float(account_dict['balance'])

        positions_dict['USD'] = utils.AttributeDict({
            'available': cash(),
            'hold': 0.0
        })

        # Now query all open orders and accordingly adjust
        open_orders = self.calls.get_all_open_orders()['orders']
        for position in open_orders:
            # todo: handle other types of orders
            if position['type'] == 'LIMIT':
                instrument_sides = position['instrument'].split('_')
                if float(position['units']) > 0:
                    quote_currency = instrument_sides[1]
                    positions_dict[quote_currency]['available'] -= float(position['units']) * float(position['price'])
                    positions_dict[quote_currency]['hold'] += float(position['units']) * float(position['price'])
                else:
                    base_currency = instrument_sides[0]
                    # sell orders just have a negative 'units'
                    positions_dict[base_currency]['available'] -= (-1 * float(position['units']))
                    positions_dict[base_currency]['hold'] += (-1 * float(position['units']))

        # Note that now __unique assets could be uninitialized:
        if self.unique_assets is None:
            self.init_exchange()

        for i in self.unique_assets:
            if i not in positions_dict:
                positions_dict[i] = utils.AttributeDict({
                    'available': 0.0,
                    'hold': 0.0
                })

        if symbol is not None:
            if symbol in positions_dict:
                return utils.AttributeDict({
                    'available': positions_dict[symbol]['available'],
                    'hold': positions_dict[symbol]['hold']
                })
            else:
                raise KeyError('Symbol not found.')

        return positions_dict

    # funds is the base asset (EUR_CAD the base asset is CAD)
    @utils.order_protection
    def market_order(self, symbol: str, side: str, size: float) -> MarketOrder:
        symbol = self.__convert_blankly_to_oanda(symbol)

        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))

        # Make sure that default trunc has been established - init may have been skipped
        if self.default_trunc is None:
            self.init_exchange()

        qty_to_buy = size
        qty_to_buy = utils.trunc(qty_to_buy, self.default_trunc)

        # we need to round qty_to_buy

        if side == "buy":
            pass
        elif side == "sell":
            qty_to_buy *= -1
        else:
            raise ValueError("side needs to be either sell or buy")

        resp = self.calls.place_market_order(symbol, qty_to_buy)

        needed = self.needed['market_order']
        order = {
            'size': qty_to_buy,
            'side': side,
            'symbol': symbol,
            'type': 'market'
        }
        if 'orderRejectTransaction' in resp:
            raise APIException(resp['errorMessage'])
        resp['symbol'] = self.__convert_symbol_to_blankly(resp['orderCreateTransaction']['instrument'])
        resp['id'] = resp['orderCreateTransaction']['id']
        resp['created_at'] = resp['orderCreateTransaction']['time']
        resp['size'] = abs(qty_to_buy)
        resp['status'] = "active"
        resp['type'] = 'market'
        resp['side'] = side

        resp = utils.isolate_specific(needed, resp)
        return MarketOrder(order, resp, self)

    @utils.order_protection
    def limit_order(self, symbol: str, side: str, price: float, size: float) -> LimitOrder:
        symbol = self.__convert_blankly_to_oanda(symbol)
        if side == "buy":
            pass
        elif side == "sell":
            size *= -1
        else:
            raise ValueError("side needs to be either sell or buy")

        resp = self.calls.place_limit_order(symbol, size, price)
        needed = self.needed['limit_order']
        order = {
            'size': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'limit'
        }

        try:
            resp['symbol'] = self.__convert_symbol_to_blankly(resp['orderCreateTransaction']['instrument'])
        except KeyError as e:
            if 'orderRejectTransaction' in resp:
                raise InvalidOrder(resp['orderRejectTransaction']['rejectReason'])
            else:
                raise e
        resp['id'] = resp['orderCreateTransaction']['id']
        resp['created_at'] = resp['orderCreateTransaction']['time']
        resp['price'] = price
        resp['size'] = abs(size)
        resp['status'] = "active"
        resp['time_in_force'] = 'GTC'
        resp['type'] = 'limit'
        resp['side'] = side

        resp = utils.isolate_specific(needed, resp)
        return LimitOrder(order, resp, self)

    def cancel_order(self, symbol, order_id) -> dict:
        # Either the Order’s OANDA-assigned OrderID or the Order’s client-provided ClientID prefixed by the “@” symbol
        self.calls.cancel_order(order_id)
        return {'order_id': order_id}

    def get_open_orders(self, symbol=None):
        if symbol is None:
            resp = self.calls.get_all_open_orders()
        else:
            symbol = self.__convert_blankly_to_oanda(symbol)
            resp = self.calls.get_orders(symbol)

        orders = resp['orders']

        for i in range(len(orders)):
            orders[i] = self.homogenize_order(orders[i])
            orders[i]['symbol'] = self.__convert_symbol_to_blankly(orders[i]['symbol'])
        return orders

    def get_order(self, symbol, order_id) -> dict:
        # Either the Order’s OANDA-assigned OrderID or the Order’s client-provided ClientID prefixed by the “@” symbol
        order = self.calls.get_order(order_id)
        homogenized = self.homogenize_order(order['order'])
        homogenized['symbol'] = self.__convert_symbol_to_blankly(homogenized['symbol'])
        homogenized['size'] = abs(homogenized['size'])
        return homogenized

    def get_fees(self, symbol):
        return {
            'maker_fee_rate': 0.0,
            'taker_fee_rate': 0.0
        }

    def overridden_history(self, symbol, epoch_start, epoch_stop, resolution, **kwargs) -> pd.DataFrame:
        resolution = self.multiples_keys[min(range(len(self.multiples_keys)),
                                             key=lambda i: abs(self.multiples_keys[i] - resolution))]
        symbol = self.__convert_blankly_to_oanda(symbol)
        if kwargs['to'] is not None:
            to = kwargs['to']

            frames = []

            while to > 0:
                # Grab candles 5k at a time
                frames.append(self.calls.get_last_k_candles(symbol, granularity=self.supported_multiples[resolution],
                                                            to_unix=epoch_stop, count=self.max_candles))

                # Step back
                epoch_stop = epoch_stop - (resolution * self.max_candles)
                to = to - self.max_candles

            # Then just trim to the number of points that are required
            return self.format_oanda_df(frames, sort=True).iloc[-kwargs['to']:].reset_index(drop=True)

        else:
            return self.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def get_product_history(self, symbol: str, epoch_start: float, epoch_stop: float, resolution: int):
        symbol = self.__convert_blankly_to_oanda(symbol)

        resolution = int(utils.time_interval_to_seconds(resolution))

        if resolution not in self.multiples_keys:
            utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
            resolution = self.multiples_keys[min(range(len(self.multiples_keys)),
                                                 key=lambda i: abs(self.multiples_keys[i] - resolution))]

        # Grab an entire window
        candles = []
        window_start = epoch_stop - (resolution * self.max_candles)
        window_end = epoch_stop
        ran_once = False
        while window_start > epoch_start or not ran_once:
            candles.append(self.calls.get_candles_by_startend(symbol,
                                                              self.supported_multiples[resolution],
                                                              window_start,
                                                              window_end))

            window_start = window_start - (resolution * self.max_candles)
            window_end = window_end - (resolution * self.max_candles)

            ran_once = True

        df = self.format_oanda_df(candles, sort=True)
        return df[df['time'] >= epoch_start].reset_index(drop=True)

    def get_order_filter(self, symbol: str):
        symbol = self.__convert_blankly_to_oanda(symbol)
        resp = self.calls.get_account_instruments(symbol)['instruments'][0]
        currencies = resp['name'].split('_')
        resp['symbol'] = self.__convert_symbol_to_blankly(resp.pop('name'))
        resp['base_asset'] = currencies[0]
        resp['quote_asset'] = currencies[1]
        resp['limit_order'] = {
            "base_min_size": float(resp['minimumTradeSize']),
            "base_max_size": float(resp['maximumOrderUnits']),
            "base_increment": 10 ** (-1 * int(resp['tradeUnitsPrecision'])),
            "price_increment": 0.01,
            "min_price": 0.00001,
            "max_price": 9999999999
        }
        price = self.get_price(symbol)
        resp['market_order'] = {
            "fractionable": True,

            "base_min_size": float(resp['minimumTradeSize']),  # Minimum size to buy
            "base_max_size": float(resp['maximumOrderUnits']),  # Maximum size to buy
            "base_increment": 10 ** (-1 * int(resp['tradeUnitsPrecision'])),  # Specifies the minimum increment

            "quote_increment": 0.01,
            "buy": {
                "min_funds": float(resp['minimumTradeSize']) * price,
                "max_funds": float(resp['maximumOrderUnits']) * price
            },
            "sell": {
                "min_funds": float(resp['minimumTradeSize']) * price,
                "max_funds": float(resp['maximumOrderUnits']) * price
            }
        }
        resp['max_orders'] = 10000
        needed = self.needed["get_order_filter"]
        resp = utils.isolate_specific(needed, resp)
        return resp

    def get_price(self, symbol: str) -> float:
        symbol = self.__convert_blankly_to_oanda(symbol)
        resp = self.calls.get_order_book(symbol)
        if 'errorMessage' in resp:
            # could be that this is instrument without order book so try using latest candle
            resp2 = self.calls.get_last_k_candles(symbol, 'S5', time.time(), 1)
            if 'errorMessage' in resp2:
                raise APIException(f'{symbol} did not have orderbook, so tried to use latest candle to '
                                   f'price. Here is the orderbook error ' + str(resp) + 'here is the candle error ' +
                                   str(resp2))
            return float(resp2['candles'][0]['mid']['c'])
        return float(resp['orderBook']['price'])

    def homogenize_order(self, order):
        def add_details(order_: dict):
            if float(order_['units']) < 0:
                order_['side'] = 'sell'
            else:
                order_['side'] = 'buy'
            order_['time_in_force'] = 'GTC'
            renames_ = [['instrument', 'symbol'],
                        ['createTime', 'created_at'],
                        ['state', 'status'],  # TODO: handle status
                        ['units', 'size']]
            return utils.rename_to(renames_, order_)
        if order['type'] == "MARKET":
            order['type'] = 'market'
            order = add_details(order)

        elif order['type'] == "LIMIT":
            order['type'] = 'limit'
            order = add_details(order)
        else:
            # TODO: handle other order types
            pass

        needed = self.choose_order_specificity(order['type'])
        order = utils.isolate_specific(needed, order)
        return order

    @staticmethod
    def __evaluate_multiples(valid_resolutions, resolution_seconds):
        found_multiple = -1
        for multiple in reversed(valid_resolutions):
            if resolution_seconds % multiple == 0:
                found_multiple = multiple
                break

        row_divisor = resolution_seconds / found_multiple
        row_divisor = int(row_divisor)

        if row_divisor > 100:
            raise Warning("The resolution you requested is an extremely high of the base resolutions supported and may "
                          "slow down the performance of your model: {} * {}".format(found_multiple, row_divisor))

        return found_multiple, row_divisor

    @staticmethod
    def _handle_input_time_conv(date: Union[str, dt, float] = None) -> float:
        if not date:
            return time.time()

        if isinstance(date, float):
            pass
        elif isinstance(date, str):
            date = dp.parse(date).timestamp()
        elif isinstance(date, dt):
            date = date.timestamp()
        else:
            raise ValueError("start and end time must be one of [str, float, datetime]")

        return date

    def _get_default_truncation(self) -> int:
        instruments = self.calls.get_account_instruments()['instruments']
        lowest_precision = int(instruments[0]['tradeUnitsPrecision'])
        for instrument in instruments:
            lowest_precision = min(lowest_precision, int(instrument['tradeUnitsPrecision']))
        return lowest_precision

    @staticmethod
    def __convert_symbol_to_blankly(symbol: str) -> str:
        return symbol.replace('_', '-')

    @staticmethod
    def __convert_blankly_to_oanda(symbol: str) -> str:
        return symbol.replace('-', '_')

    @staticmethod
    def format_oanda_df(candles: list, sort: bool = False):
        """
        Given an OANDA history type response, format it into a readable dataframe:
        Example index from a list response:

        (this is a 1 element list with a single query)
        [{'instrument': 'EUR_USD', 'granularity': 'H1', 'candles':
            [{'complete': True, 'volume': 1202, 'time': '1637114400.000000000', 'mid': {'o': '1.13232', 'h': '1.13242',
                'l': '1.13189', 'c': '1.13218'}},
            {'complete': True, 'volume': 4907, 'time': '1637118000.000000000', 'mid': {'o': '1.13216', 'h': '1.13220',
                'l': '1.12637', 'c': '1.12890'}},
            {'complete': True, 'volume': 2478, 'time': '1637121600.000000000', 'mid': {'o': '1.12891', 'h': '1.13015',
                'l': '1.12890', 'c': '1.12982'}}]
        }]

        This gets kind of weird because OANDA gives a list of responses:
        [resp1, resp2] where each response *might* have a candles key and inside a candles key
        is the actual data we need
        """
        result = []
        for single_response in candles:
            try:
                internal_candles = single_response['candles']
            except KeyError:
                # There must not be a candles key to even get to the mid.
                # This will generate an empty df
                internal_candles = []

            for single_candle in internal_candles:
                ohlc = single_candle['mid']
                result.append([int(float(single_candle['time'])),
                               ohlc['o'],
                               ohlc['h'],
                               ohlc['l'],
                               ohlc['c'],
                               single_candle['volume']])

        df = pd.DataFrame(result, columns=['time', 'open', 'high', 'low', 'close', 'volume'])
        dtypes = {"time": int, "open": float, "high": float, "low": float, "close": float,
                  "volume": float}
        if sort:
            return df.sort_values(by=['time'], ignore_index=True).astype(dtypes)
        else:
            return df.astype(dtypes)
